"""Definition of the ReLeSO base Model.

File holding the base class for all ReLeSO classes which are needed for the
command line based application of this toolbox.
"""
import multiprocessing
import pathlib
from collections import OrderedDict
from typing import Any, Dict, List, Optional

import pydantic

from releso.util.logger import logging
from releso.util.util_funcs import get_path_extension


def add_save_location_if_elem_is_o_dict(
    possible_value: Any, save_location: pathlib.Path
):
    """Add the save_location keyword to the element of it is a dict.

    This function adds the save_location to all ordered dicts in possible_value
    if it is not already present. This function is used to forward the
    save_location variable to all child objects so that all objects can access
    it.

    Also handles adding the save_location to objects inside of lists (these can
    even be multiple lists deep)

    Args:
        possible_value (Any): Variable holding potential objects which need
        the save_location.
        save_location (str): string that defines the location
    """
    if isinstance(possible_value, OrderedDict) or isinstance(
        possible_value, Dict
    ):
        if "save_location" not in possible_value:
            possible_value["save_location"] = save_location
    elif isinstance(possible_value, list):
        for item in possible_value:
            add_save_location_if_elem_is_o_dict(item, save_location)
    else:
        pass


class BaseModel(pydantic.BaseModel):
    """Base of all ReLeSO objects used for parsing.

    Base class for all ReLeSO classes which are needed for the command line
    based application of this toolbox.
    """

    #: Definition of the save location of the
    #: logs and validation results. Should be given as a standard string will
    #: be pre-converted into a pathlib.Path. If {} is present in the string the
    #: current timestamp is added if in a slurm job SLURM_JOB_ID are added.
    save_location: pathlib.Path
    #: name of the logger. If this variable gives you trouble, the framework is
    #: at least a little bit buggy. Does not need to be set. And might be
    #: changed if set by user, if multi-environment training is utilized.
    logger_name: Optional[str] = None

    def __init__(self, **data: Any) -> None:
        """Constructor for the ReLeSO basemodel object."""
        if "save_location" in data:
            # This is so that the save location always gets the same
            if isinstance(data["save_location"], str):
                data[
                    "save_location"
                ] = BaseModel.convert_to_pathlib_add_datetime(
                    data["save_location"]
                )
            # if a save_location is present in the current object definition
            #  add this save_location also to all object definition which are
            #  direct dependents
            for value in data.values():
                add_save_location_if_elem_is_o_dict(
                    value, data["save_location"]
                )
        super().__init__(**data)

    @classmethod
    def convert_to_pathlib_add_datetime(cls, v: str):
        """Add timestamp to save_location, of applicable.

        Adds a datetime timestamp to the save_location if {} present and
        current slurm job id if available. Task arrays will be added to the
        same folder and each task gets its own subfolder. This is done to make
        it easier to differentiate different runs without the need to change
        the base name with every run and to easily group similar runs.

        This function also ensures that the directory is/directories are
        created.

        Args:
            v ([type]): Value to validate

        Returns:
            pathlib.Path: Path like object. If applicable with identifications.
        """
        path = (
            pathlib.Path(v.format(get_path_extension())).expanduser().resolve()
        )
        path.mkdir(parents=True, exist_ok=True)
        return path

    def _check_list(self, list_item: List[Any], logger_name: str):
        """Helper function for set_logger_name_recursively.

        Recursively goes through lists and add the logger_name where
        applicable.

        Args:
            list_item (List[Any]): list in which to check if items reside
            which need checking for logger names
            logger_name (str): logger name to add to all applicable objects
        """
        for item in list_item:
            if isinstance(item, list):
                self._check_list(item, logger_name)
            elif isinstance(item, BaseModel):
                item.set_logger_name_recursively(logger_name)

    def set_logger_name_recursively(self, logger_name: str):
        """Set the logger_name variable for all child elements.

        Sets the given logger_name for the current object and all attributes
        which have the ``set_logger_recursively`` method.

        Note:
            Please note: It is the callings functions responsibility to ensure
            that the logger actually exists.

        Args:
            logger_name (str): Name of the logger to set.
        """
        self.logger_name = logger_name
        for value in self.__dict__.values():
            if isinstance(value, list):
                self._check_list(value, logger_name)
            else:
                attr = getattr(value, "set_logger_name_recursively", None)
                if callable(attr):
                    attr(logger_name)

    def get_logger(self) -> logging.Logger:
        """Gets the currently defined environment logger.

        If multiprocessing is part of the logger name the multiprocessing
        standard logger will be called.

        Returns:
            logging.Logger: logger which is currently to be used.
        """
        if self.logger_name and "multiprocessing" in self.logger_name:
            return multiprocessing.get_logger()
        return logging.getLogger(self.logger_name)

    # makes it that pydantic returns an error if unknown keywords are given
    class Config:
        """Used to add pydantic configurations."""

        #: Forbids superfluous keywords for object definitions.
        extra = pydantic.Extra.forbid



#!/usr/bin/env python
"""ReLeSO main file and framework starting point.

File defines the main entry point of the framework if it is called via the
command line. (via $python -m ReLeSO; or $releso)
"""

import argparse
import datetime
import pathlib
import pprint
import shutil
from typing import Literal, Union

import gymnasium
import hjson
import numpy as np
import stable_baselines3
import torch

from releso.__version__ import __version__
from releso.base_parser import BaseParser
from releso.util.module_import_raiser import ModuleImportRaiser

try:
    import splinepy
except ModuleNotFoundError:
    splinepy = ModuleImportRaiser("splinepy")

try:
    from releso.util.visualization import (
        export_figure,
        plot_episode_log,
        plot_step_log,
    )
except ModuleNotFoundError:
    export_figure = ModuleImportRaiser("plotly")
    plot_episode_log = export_figure
    plot_step_log = export_figure


def check_positive(value) -> int:
    """Checks if the provided value is positive.

    Note:
        Adapted from https://stackoverflow.com/a/14117511.

    Args:
        value (Any): Value to be cast to integer.

    Raises:
        argparse.ArgumentTypeError: Throws error if the value is not positive.

    Returns:
        int: checked value
    """
    value = int(value)
    if value <= 0:
        raise argparse.ArgumentTypeError("Provided value is not positive.")
    return value


def check_positive_or_zero(value) -> int:
    """Checks if the provided value is positive or zero.


    Args:
        value (Any): Value to be cast to integer.

    Raises:
        argparse.ArgumentTypeError: Throws error if the value is not positive
            or zero.

    Returns:
        int: checked value
    """
    value = int(value)
    if value < 0:
        raise argparse.ArgumentTypeError(
            "Provided value is not positive or zero."
        )
    return value


def check_window_size(value) -> Union[tuple[int, int], Literal["auto"]]:
    """Check and validate window size.
    Args:
        value (Any): Value to be validated. Can be "auto" or a tuple of two integers.
    Raises:
        argparse.ArgumentTypeError: If value is not correct.
    Returns:
        Union[tuple[int, int], Literal["auto"]]: Validated value.
    """
    if str(value) == "auto":
        return "auto"
    try:
        # Check if value is a tuple or string representation of a tuple
        if isinstance(value, tuple) and len(value) == 2:
            width, height = map(int, value)
            if width > 0 and height > 0:
                return (width, height)
        elif (
            isinstance(value, str)
            and value.startswith("(")
            and value.endswith(")")
        ):
            width, height = map(int, value.strip("()").split(","))
            if width > 0 and height > 0:
                return (width, height)
    except (ValueError, TypeError):
        pass
    raise argparse.ArgumentTypeError(
        "Provided value must be 'auto' or a tuple of two positive integers."
    )


def main(args) -> pathlib.Path:
    """Calling function of framework.

    Functions control how the framework works when called from the command
    line.

    Args:
        args ([type]): Command line arguments

    Raises:
        ValueError: Thrown if the json file could not be found.
    """
    ###########################
    #                         #
    #   Loading and parsing   #
    #     the json file       #
    #                         #
    ###########################
    file_path = pathlib.Path(args.input_file).expanduser().resolve()
    if file_path.exists() and file_path.is_file():
        with open(file_path) as file:
            json_content = hjson.load(file)
    else:
        raise ValueError(f"Could not find the given file: {file_path}.")

    optimization_object = BaseParser(**json_content)

    ###########################
    #                         #
    #  Save json definition   #
    #                         #
    ###########################

    shutil.copy(file_path, optimization_object.save_location / file_path.name)
    versions = [
        f"releso version: {__version__}; ",
        f"stable-baselines3 version: {stable_baselines3.__version__}; ",
        f"torch version: {torch.__version__}; ",
        f"gymnasium version: {gymnasium.__version__} ",
    ]
    if splinepy:
        versions.append(f"splinepy version: {splinepy.__version__}")
    optimization_object.get_logger().info("".join(versions))
    ###########################
    #                         #
    #   Validate json file    #
    #                         #
    ###########################

    if args.json_validate:
        pprint.pprint(optimization_object.dict())
        return

    ###########################
    #                         #
    #    Only validation      #
    #                         #
    ###########################
    if args.validate_only:
        optimization_object.evaluate_model(throw_error_if_none=True)
        return

    ###########################
    #                         #
    #  Training the RL Agent  #
    #                         #
    ###########################
    optimization_object.learn()

    return optimization_object.save_location


def entry():
    """Entry point if this package is called directly from the command line."""
    parser = argparse.ArgumentParser(
        description=(
            "Reinforcement Learning based Shape Optimization (releso) Toolbox. "
            "This python program loads a problem definition and trains an RL "
            "agent to solve the resulting problem. Furthermore a trained "
            "agent can be evaluated or data gathered during a training "
            f"visualized. The package version is: {__version__}."
        )
    )
    parser.add_argument(
        "--version",
        dest="version",
        action="store_true",
        help="Returns the version of the package.",
    )
    sub_parser = parser.add_subparsers(
        title="execution modes", dest="execution_mode", required=True
    )
    parser_run = sub_parser.add_parser(
        "run",
        help=(
            "Start a releso run with a given json file. Choose this option if "
            "you want to train a model or validate a given input file."
        ),
    )
    parser_run.add_argument(
        "-i",
        "--input-file",
        action="store",
        required=True,
        help="Path to the json file storing the optimization definition.",
    )
    parser_run.add_argument(
        "-v",
        "--validate-only",
        action="store_true",
        help=(
            "If this is set only validation on this configuration is run. "
            "Please configure the validation object in the json file so that "
            "this option can be correctly executed.",
        ),
    )
    parser_run.add_argument(
        "-j",
        "--json-only",
        dest="json_validate",
        action="store_true",
        help=(
            "If this is set only the json validation is performed, nothing "
            "else."
        ),
    )
    parser_visualize = sub_parser.add_parser(
        "visualize",
        help=(
            "Visualize training data. Choose between visualizing the data "
            "contained in the episode log (global training information) and "
            "the data contained in the step log (more in-depth training "
            "specific information)."
        ),
    )
    visualize_shared_args = argparse.ArgumentParser(add_help=False)
    visualize_shared_args.add_argument(
        "-e",
        "--export-path",
        type=pathlib.Path,
        default="./",
        help=(
            "Path where the visualization should be saved to. If a directory "
            "is provided the created figure will be saved in that directory "
            "as html file under a default filename which depends on the type "
            "of visualization that has been chosen. If the path contains a "
            "filename with a suffix, it will be exported as the file type "
            "indicated by the suffix if that is possible. Available suffixes "
            "are [png, jpg, webp, svg, pdf, html]. Defaults to './'."
        ),
    )
    visualize_shared_args.add_argument(
        "-s",
        "--figure-size",
        type=check_window_size,
        default="auto",
        nargs="+",
        help="Size of the figure.",
    )
    sub_parser_visualize = parser_visualize.add_subparsers(
        title="visualization modes", dest="visualization_mode", required=True
    )
    parser_visualize_episodelog = sub_parser_visualize.add_parser(
        "episode-log",
        parents=[visualize_shared_args],
        help=(
            "Visualize training progress. List all folders you want to "
            "include. If run a training during this call as well the visualization "
            "will be exported after the training has finished and will also "
            "include the new results automatically. The legend name of each "
            "experiment is the folder name. If you need to display custom "
            "names please use the call the function from your own script. "
            "You can find the function in 'releso.util.visualization.plot_episode_log'. "
        ),
    )
    parser_visualize_episodelog.add_argument(
        "-f",
        "--folders",
        type=pathlib.Path,
        nargs="*",
        required=True,
        help=(
            "Visualize given training episode logs. You can also use wildcard "
            "arguments '*' or '?' at least on some systems."
        ),
    )
    parser_visualize_episodelog.add_argument(
        "-w",
        "--window",
        type=check_positive,
        default=5,
        help=(
            "Episode visualization uses windowing to smooth the graph. Set "
            "the window length. Defaults to 5."
        ),
    )
    parser_visualize_episodelog.add_argument(
        "-c",
        "--cut-off-point",
        type=check_positive,
        default=np.iinfo(int).max,
        help=(
            "Cut off point for the visualization (in timesteps). If set, the "
            "visualization will exclude all timesteps after the specified "
            "cutoff point. Defaults to np.iinfo(int).max, which means that all "
            "timesteps will be included."
        ),
    )
    parser_visualize_steplog = sub_parser_visualize.add_parser(
        "step-log",
        parents=[visualize_shared_args],
        help=(
            "Visualize the strategy of the agent employed in each episode. "
            "If you want to customize the visualization please use the call "
            "the corresponding function from your own script. "
            "You can find the function in 'releso.util.visualization.plot_step_log'. "
        ),
    )
    parser_visualize_steplog.add_argument(
        "-l",
        "--logfile",
        type=pathlib.Path,
        required=True,
        help=(
            "Visualize a given training step log to analyze the strategy "
            "learned by an agent in a specific run."
        ),
    )
    parser_visualize_steplog.add_argument(
        "-i",
        "--environment-id",
        type=check_positive_or_zero,
        default=0,
        help=(
            "ID of the environment whose data is supposed to be visualized "
            "(only relevant for multi-environment trainings). Defaults to 0."
        ),
    )
    parser_visualize_steplog.add_argument(
        "-f",
        "--from-episode",
        type=check_positive_or_zero,
        default=0,
        help=(
            "Starting episode of the interactive visualization. Defaults to 0."
        ),
    )
    parser_visualize_steplog.add_argument(
        "-n",
        "--n-episodes",
        type=check_positive,
        default=1,
        help=(
            "Select the amount of episodes skipped between visualized episodes."
            " Defaults to 1 which means that every episode between from_episode"
            " and until_episode will be visualized. Set to a larger value to "
            "include less episodes in the visualization."
        ),
    )
    parser_visualize_steplog.add_argument(
        "-u",
        "--until-episode",
        type=check_positive,
        default=None,
        help=(
            "Final episode of the interactive visualization. Defaults to None, "
            "which means that all episodes after the chosen starting one will "
            "be visualized."
        ),
    )
    args = parser.parse_args()
    if args.version:
        print(f"releso: {__version__}")
        return

    run_folder = None

    # We want to run releso
    if args.execution_mode == "run":
        print(datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S"))
        run_folder = main(args)
    # We want to use releso for visualization
    elif args.execution_mode == "visualize":
        # Determine the figure size
        figure_size = args.figure_size
        if len(figure_size) == 1:
            figure_size = figure_size[0]
        # Visualize contents of episode_log.csv
        if args.visualization_mode == "episode-log":
            folders_to_process: list = [
                folder.resolve() for folder in args.folders
            ]
            if run_folder is not None:
                folders_to_process.append(run_folder)
            folder_dict = {
                folder.stem: folder for folder in folders_to_process
            }

            fig = plot_episode_log(
                folder_dict,
                args.window,
                window_size=figure_size,
                cut_off_point=args.cut_off_point,
            )
            export_figure(fig, args.export_path, "episode_log.html")
        # Visualize contents of step_log.jsonl
        elif args.visualization_mode == "step-log":
            fig = plot_step_log(
                args.logfile.resolve(),
                args.environment_id,
                episode_start=args.from_episode,
                episode_end=args.until_episode,
                episode_step=args.n_episodes,
                figure_size=figure_size,
            )
            # Export the plot as the suffix or as "steplog_plot.html"
            export_figure(fig, args.export_path, "steplog_plot.html")
        else:
            print(
                f"Unknown visualization mode: {args.visualization_mode}. "
                "Please choose between 'episode-log' and 'step-log'."
            )
            parser.print_help()
            exit(1)
    else:
        parser.print_help()
        exit(1)


if __name__ == "__main__":  # pragma: no cover
    entry()



import os
import pathlib
import shutil

import pytest
from pydantic import ValidationError

from releso.mesh import MeshExporter, MeshIOMesh, MixdMesh

# will look into gustaf test meshes to test these classes


@pytest.mark.parametrize(
    "path, mesh_format, expected, error",
    [
        ("test", "test", False, "mesh_format='test' is not supported"),
        (
            "test.txt",
            "mixd",
            False,
            "The format and the path suffix do not match. The format",
        ),
        ("test.xns", "mixd", pathlib.Path("test.xns").resolve(), None),
        ("test", "mixd", pathlib.Path("_.xns").resolve(), None),
        ("test/test", "mixd", pathlib.Path("test/_.xns").resolve(), None),
    ],
)
def test_mesh_exporter_export_path_format(
    path, mesh_format, expected, error, dir_save_location
):
    calling_dict = {
        "export_path": path,
        "format": mesh_format,
        "save_location": dir_save_location,
    }
    if error:
        with pytest.raises(ValidationError) as err:
            MeshExporter(**calling_dict)
        assert error in str(err.value)
        return
    mesh_exporter = MeshExporter(**calling_dict)
    assert mesh_exporter.export_path == expected
    temp_expected = pathlib.Path(str(expected).format(0))
    assert mesh_exporter.get_export_path() == temp_expected
    env_id = "123"
    mesh_exporter.adapt_export_path(env_id)
    expected = pathlib.Path(str(expected).format(env_id))
    assert mesh_exporter.get_export_path() == expected
    if "/" in path:
        os.rmdir(expected.parent)


@pytest.mark.parametrize(
    "load_sample_file",
    [
        "volumes/tet/3DBrickTet.msh",
    ],
    indirect=True,
)
@pytest.mark.parametrize(
    (
        "path",
        "mxyz_path",
        "mien_path",
        "hyper_cube",
        "export",
        "dimensions",
        "w_level",
        "warning",
        "error",
    ),
    [
        (
            ".xns",
            None,
            None,
            False,
            {"format": "mixd", "export_path": "test"},
            3,
            False,
            False,
            False,
        ),
        (
            ".xns",
            None,
            None,
            False,
            None,
            3,
            "export",
            "No mesh exporter definition. Please define one",
            False,
        ),
        (
            "mxyz",
            None,
            None,
            False,
            {"format": "mixd", "export_path": "test"},
            3,
            False,
            False,
            False,
        ),
        (
            "mien",
            None,
            None,
            False,
            {"format": "mixd", "export_path": "test"},
            3,
            False,
            False,
            False,
        ),
        (
            "mxyz_no_name",
            None,
            None,
            False,
            {"format": "mixd", "export_path": "test"},
            3,
            False,
            False,
            False,
        ),
        (
            "mien_no_name",
            None,
            None,
            False,
            {"format": "mixd", "export_path": "test"},
            3,
            False,
            False,
            False,
        ),
        (
            None,
            "correct",
            "correct",
            False,
            {"format": "mixd", "export_path": "test"},
            3,
            False,
            False,
            False,
        ),
        (
            "msh",
            None,
            None,
            False,
            {"format": "mixd", "export_path": "test"},
            3,
            False,
            False,
            "Could not locate mien nor mxyz file path.",
        ),
        (
            None,
            None,
            None,
            False,
            {"format": "mixd", "export_path": "test"},
            3,
            False,
            False,
            "Could not locate the correct mien/mxyz paths.",
        ),
    ],
)
def test_mixd_mesh_class(
    path,
    mxyz_path,
    mien_path,
    hyper_cube,
    export,
    dimensions,
    w_level,
    warning,
    error,
    load_sample_file,
    dir_save_location,
    caplog,
):
    # setup mixd
    if not load_sample_file.with_suffix(".mien").exists():
        from gustaf.io import load
        from gustaf.io.mixd import export as export_io

        mesh = load(load_sample_file)[-1]
        export_io(load_sample_file.with_suffix(".xns"), mesh)
        shutil.copy(
            load_sample_file.with_suffix(".mxyz"),
            load_sample_file.parent / "mxyz",
        )
        shutil.copy(
            load_sample_file.with_suffix(".mxyz"),
            load_sample_file.parent / "mxyz.space",
        )
        shutil.copy(
            load_sample_file.with_suffix(".mien"),
            load_sample_file.parent / "mien",
        )
    load_sample_file = load_sample_file.with_suffix(".xns")
    calling_dict = {
        "save_location": dir_save_location,
    }
    if path:
        if path == ".xns":
            path = load_sample_file.parent
        elif path == "mxyz":
            path = load_sample_file.with_suffix(".mxyz")
        elif path == "mien":
            path = load_sample_file.with_suffix(".mien")
        elif path == "mxyz_no_name":
            path = load_sample_file.with_name("mxyz")
        elif path == "mien_no_name":
            path = load_sample_file.with_name("mien")
        elif path == "msh":
            path = load_sample_file.with_suffix(".msh")
        calling_dict["path"] = path
    if mxyz_path:
        if mxyz_path == "correct":
            mxyz_path = load_sample_file.with_suffix(".mxyz")
        calling_dict["mxyz_path"] = mxyz_path
    if mien_path:
        if mien_path == "correct":
            mien_path = load_sample_file.with_suffix(".mien")
        calling_dict["mien_path"] = mien_path
    if hyper_cube:
        calling_dict["hyper_cube"] = hyper_cube
    if dimensions:
        calling_dict["dimensions"] = dimensions
    if export:
        export["save_location"] = dir_save_location
        calling_dict["export"] = MeshExporter(**export)
    if error:
        with pytest.raises(ValidationError) as err:
            MixdMesh(**calling_dict)
        assert error in str(err.value)
        return
    mesh = MixdMesh(**calling_dict)
    if w_level == "export":
        assert mesh.get_export_path() is None
        with caplog.at_level("WARNING"):
            assert mesh.adapt_export_path("123") is None
            assert warning in caplog.text
    else:
        assert mesh.get_export_path() is not None


@pytest.mark.parametrize(
    "load_sample_file, dimension, change_path, cp_rm, error",
    [
        ("volumes/tet/3DBrickTet.msh", 3, False, False, False),
        # currently not supported
        # ("volumes/tet/3DBrickTet.msh", ".msh2", True, False),
        (
            "volumes/tet/3DBrickTet.msh",
            3,
            ".msh2",
            False,
            "Could not locate the mesh file.",
        ),
        (
            "volumes/tet/3DBrickTet.msh",
            3,
            ".msh3",
            True,
            "Mesh type not supported.",
        ),
        (
            "volumes/tet/3DBrickTet.msh",
            3,
            "rm",
            False,
            "For MeshIO base mesh import a path must be provided.",
        ),
        ("faces/quad/2DChannelQuad.msh", 2, False, False, False),
    ],
    indirect=["load_sample_file"],
)
def test_meshio_mesh(
    load_sample_file, dimension, change_path, cp_rm, error, dir_save_location
):
    if change_path:
        if change_path == "rm":
            load_sample_file = None
        else:
            load_sample_file = load_sample_file.with_suffix(change_path)
            if cp_rm:
                shutil.copy(
                    load_sample_file.with_suffix(".msh"),
                    load_sample_file,
                )
    call_dict = {
        "path": load_sample_file,
        "dimensions": dimension,
        "save_location": dir_save_location,
    }
    if error:
        with pytest.raises(ValidationError) as err:
            MeshIOMesh(**call_dict)
        assert error in str(err.value)
        if change_path and cp_rm:
            os.remove(load_sample_file)
        return
    mesh = MeshIOMesh(**call_dict)
    if change_path and cp_rm:
        os.remove(load_sample_file)
    mesh.get_mesh()


@pytest.mark.parametrize(
    "load_sample_file", ["faces/quad/2DChannelQuad.msh"], indirect=True
)
def test_mixd_mesh_export_and_get(
    load_sample_file, dir_save_location, clean_up_provider, caplog
):
    mesh_exporter = MeshExporter(
        format="mixd",
        export_path=f"{str(dir_save_location)}/export/" + "{}/test.xns",
        save_location=dir_save_location,
    )
    mesh = MeshIOMesh(
        path=load_sample_file,
        dimensions=2,
        save_location=dir_save_location,
        export=mesh_exporter,
    )
    mesh.adapt_export_path("123")
    mesh.export.export_mesh(mesh.get_mesh())
    exported_file = pathlib.Path(f"{str(dir_save_location)}/export")
    mesh_exporter.mesh_format = "other"
    with pytest.raises(RuntimeError) as err:
        mesh_exporter.export_mesh(None)
    assert "The requested format other is not supported." in str(err.value)

    mixd_mesh = MixdMesh(
        path="export/123/test.xns",
        save_location=dir_save_location,
        dimensions=2,
    )
    with caplog.at_level("INFO"):
        mixd_mesh.get_mesh()

    clean_up_provider(exported_file)



import datetime
import pathlib

import pytest

from releso.util.logger import logging
from releso.verbosity import Verbosity, VerbosityLevel


def test_clean_dir(dir_save_location, clean_up_provider):
    clean_up_provider(dir_save_location)


@pytest.mark.parametrize(
    [
        "parser",
        "wanted_parser",
        "environment",
        "wanted_environment",
        "logfile_location",
        "console_logging",
        "base_logger_name",
        "environment_extension",
        "error",
    ],
    [
        (
            None,
            VerbosityLevel.INFO,
            None,
            VerbosityLevel.INFO,
            None,
            None,
            None,
            None,
            False,
        ),
        (
            "DEBUG",
            VerbosityLevel.DEBUG,
            "WARNING",
            VerbosityLevel.WARNING,
            None,
            None,
            None,
            None,
            False,
        ),
        (
            "ERROR",
            VerbosityLevel.ERROR,
            "INFO",
            VerbosityLevel.INFO,
            None,
            None,
            None,
            None,
            False,
        ),
        # ("Error", "Error", False, False, False, False),
    ],
)
def test_verbosity_default(
    parser,
    wanted_parser,
    environment,
    wanted_environment,
    logfile_location,
    console_logging,
    base_logger_name,
    environment_extension,
    error,
    dir_save_location,
    caplog,
    clean_up_provider,
):
    # check folder empty
    files_in_folder = list(dir_save_location.glob("*"))
    assert len(list(files_in_folder)) == 0

    calling_dict = {
        "save_location": dir_save_location,
    }
    if parser:
        calling_dict["parser"] = parser
    if environment:
        calling_dict["environment"] = environment
    if logfile_location:
        calling_dict["logfile_location"] = logfile_location
    if console_logging:
        calling_dict["console_logging"] = console_logging
    if base_logger_name:
        calling_dict["base_logger_name"] = base_logger_name
    if environment_extension:
        calling_dict["environment_extension"] = environment_extension
    if wanted_parser <= VerbosityLevel.DEBUG:
        with caplog.at_level(VerbosityLevel.DEBUG):
            verbosity = Verbosity(**calling_dict)
            assert (
                f"Parser logger has logging level: {str(wanted_parser.value)}"
                in caplog.text
            )
            assert (
                f"Environment logger has logging level: "
                f"{str(wanted_environment.value)}"
            ) in caplog.text
    else:
        verbosity = Verbosity(**calling_dict)

    assert verbosity.parser is wanted_parser
    assert verbosity.environment is wanted_environment

    logging.getLogger(verbosity._environment_validation_logger).warning("test")
    logging.getLogger(verbosity._environment_logger).warning("test")
    logging.getLogger("releso_parser").warning("test")
    # verbosity._parser_logger.debug("test")
    # objects_in_folder = list((dir_save_location / "logging").glob("*"))
    # # try:
    # #     assert len(objects_in_folder) == 4
    # # except AssertionError:
    # #     raise RuntimeError(f"{objects_in_folder}")
    clean_up_provider(dir_save_location)


@pytest.mark.parametrize(
    "add_save_path, log_path",
    [(True, "test"), (False, "test"), (False, "test_{}"), (True, "test_{}")],
)
def test_verbosity_make_logfile_location_absolute(
    add_save_path, log_path, clean_up_provider, dir_save_location
):
    v = log_path

    values = {}
    if add_save_path:
        values["save_location"] = dir_save_location
    ret_path = Verbosity.make_logfile_location_absolute(v, values)
    if add_save_path:
        assert ret_path == dir_save_location / v.format(
            datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        )
    else:
        if "{}" in v:
            assert (
                ret_path
                == pathlib.Path(
                    v.format(
                        datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
                    )
                ).resolve()
            )
        else:
            assert ret_path == pathlib.Path(v).resolve()
    clean_up_provider(ret_path)



"""SPOR step python file making cnn observations based on xns results.

This file is not tested due to it being specific to the xns use case.
Please check if any updates to the required files are needed before
using these functions.
"""
import datetime
import pathlib
from typing import Any, Dict, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np

from releso.observation import ObservationDefinitionMulti
from releso.util.load_binary import read_mixd_double
from releso.util.plotting import get_tricontour_solution


def define_observation_definition() -> ObservationDefinitionMulti:
    """Returns the observation definition for the SPORObject.

    This is a image based observation creation routine. This means, that the
    observation created is of type CNN and has as shape of (3, 200, 200).

    Returns:
        ObservationDefinitionMulti: _description_
    """
    return ObservationDefinitionMulti(
        save_location="",  # Dummy safe location
        name="cnn_observation",
        value_min=0,
        value_max=255,
        observation_shape=(3, 200, 200),
        value_type="CNN",
    )


def get_visual_representation(
    connectivity: np.ndarray,
    sol_len: int = 3,
    height: int = 10,
    width: int = 10,
    dpi: int = 20,
) -> np.ndarray:
    """Return an array representing the calculated solution.

    This function is used to create the array which is used if the
    solution/cnn representation of the base observations are selected.

    Note:
        The calculated solution can only be found if the
        :py:class:`SporObject` of the solver is called main_solver and uses
        the xns solver.
    #TODO make it broader in its application e.g. other solvers?

    Args:
        connectivity (np.ndarray): Connectivity array of the mesh.
        sol_len (int, optional): Number of variables to return per data
            point, input can be more but not less. Assumed is u,v,p.
            Defaults to 3.
        height (int, optional): Height of the image array in inches.
            Defaults to 10.
        width (int, optional): Width of the image array in inches.
            Defaults to 10.
        dpi (int, optional): DPI of the image array. Together with the
            height and width, this variable defines the shape of the
            return array shape=(height*dpi, width*dpi, sol_len).
            Defaults to 20.

    Raises:
        RuntimeError: could not find solution file, mesh is of the
            incorrect type (needs to be triangular), mesh is of the
            incorrect dimensions (needs to be dim=2)
    """
    # if self.mesh.hypercube:
    #     raise RuntimeError(
    #         "The given mesh is not a mesh with triangular entities. "
    #         "This function was designed to work with only those.")
    # if not self.mesh.dimensions == 2:
    #     raise RuntimeError(
    #         "The given mesh has a dimension unequal two. This function "
    #         "was designed to work only with meshes of dimension two.")
    solution = read_mixd_double("ins.out", 3)
    coordinates = (
        np.fromfile("mesh/mxyz", dtype=">d").astype(np.float64).reshape(-1, 2)
    )

    # Plotting and creating resulting array
    limits_max = [1, 1, 0.2e8]
    limits_min = [-1, -1, -0.2e8]

    return get_tricontour_solution(
        width,
        height,
        dpi,
        coordinates,
        connectivity,
        solution,
        sol_len,
        limits_min,
        limits_max,
    )


def save_current_solution_as_png(
    save_location: Union[pathlib.Path, str],
    include_pressure: bool = True,
    height: int = 10,
    width: int = 10,
    dpi: int = 400,
):
    """Save the current solver solution as an image at the given location.

    Additional parameters for size can be used.

    TODO additional SPORStep for this functionality?

    Args:
        save_location (Union[pathlib.Path, str]): Where the image is to
            be saved to. Please add solution specific path, else the file
            will be overwritten if multiple solution are saved.
        include_pressure (bool, optional): Not only use x-,y-velocity
            fields but also the pressure field. Defaults to True.
        height (int, optional): Self explanatory. Defaults to 10.
        width (int, optional): Self explanatory. Defaults to 10.
        dpi (int, optional): Self explanatory. Defaults to 400.
    """
    image_arr = get_visual_representation(
        sol_len=3 if include_pressure else 2,
        width=height,
        height=width,
        dpi=dpi,
    )
    meta_data_dict = {
        "Author": "Clemens Fricke",
        "Software": "ReLeSO",
        "Creation Time": str(datetime.datetime.now()),
        "Description": "This is a description",
    }
    if isinstance(save_location, str):
        save_location = pathlib.Path(save_location)
    save_location.parent.mkdir(parents=True, exist_ok=True)
    plt.imsave(save_location, image_arr, metadata=meta_data_dict)


def main(args, logger, func_data) -> Tuple[Dict[str, Any], Any]:
    """Function which is called from the spor step.

    The parameters need to conform to the SPOR_COMM interface defined by
    ReLeSO.

    Args:
        args (_type_): _description_
        logger (_type_): _description_
        func_data (_type_): _description_

    Returns:
        Set[Dict[str, Any], Any]: _description_
    """
    # first time initialization
    if func_data is None:
        func_data = {}
        func_data["connectivity"] = (
            (np.fromfile("mesh/mien", dtype=">i") - 1)
            .astype(np.int64)
            .reshape(-1, 3)
        )

    # Loading the correct data and checking if correct attributes are set.
    cnn_observation = get_visual_representation(
        connectivity=func_data["connectivity"]
    )

    return_dict = {
        "reward": 0,
        "done": False,
        "info": {},
        "observations": [cnn_observation.T],
    }

    return return_dict, func_data



"""Definition of Observation space definitions.

These are used to parse and handle the observations inside the package. The
biggest functional change is that they provide default observations. Needed
in cases where observations are depended on a previous step which errored out
and could not be completed.
"""
from typing import List, Literal, Tuple

import numpy as np
from gymnasium.spaces import Box, Space

from releso.base_model import BaseModel


class ObservationDefinition(BaseModel):
    """Definition of an Observation.

    Definition of a single Observations by providing the name of the
    value and the range in which the value is limited to.

    The range is necessary due to normalization of the input of the agent
    networks.
    """

    name: str  #: Name of the observation
    #: minimum of the range in which the observation is bound
    value_min: float
    #: maximum of the range in which the observation is bound
    value_max: float

    def get_observation_definition(self) -> Tuple[str, Space]:
        """Provide definition of the defined observations space.

        Returns a tuple of name and observation definition defined via the
        gymnasium.observation interface

        Returns:
            Tuple[str, Space]: Tuple of the name of the observation and a
            gymnasium.Box definition
        """
        return (
            self.name,
            Box(self.value_min, self.value_max, shape=([1]), dtype=np.float32),
        )

    def get_default_observation(self) -> np.ndarray:
        """Provide default observations.

        Gives a default observations of the correct shape, purpose is when the
        observation fails that the observation can still be generated so that
        the training does not error out.

        Returns:
            np.ndarray: An array filled with ones in the correct shape and size
            of the observation.
        """
        return np.ones((1,)) * self.value_min


class ObservationDefinitionMulti(ObservationDefinition):
    """Define a multidimensional Observations space.

    Definition of a single Observations by providing the name of the value
    and the range in which the value is limited to.

    The range is necessary due to normalization of the input of the agent
    networks.
    """

    #: Shape of the Observation space. List of number of elements per dimension
    observation_shape: List[int]
    #: Type of the Observation space. If float uses the value_min etc
    #: definition for the limits of the space. If CNN uses [0, 255] limits.
    value_type: Literal["float", "CNN"]

    def get_observation_definition(self) -> Tuple[str, Space]:
        """Provide definition of the defined observations space.

        Returns a tuple of name and observation definition defined via the
        gymnasium.observation interface

        Returns:
            Tuple[str, Space]: Tuple of the name of the observation
                and a gymnasium.Box definition
        """
        if self.value_type == "CNN":
            return (
                self.name,
                Box(
                    low=0,
                    high=255,
                    shape=self.observation_shape,
                    dtype=np.uint8,
                ),
            )
        return (
            self.name,
            Box(
                self.value_min,
                self.value_max,
                shape=(self.observation_shape),
                dtype=np.float32,
            ),
        )

    def get_default_observation(self) -> np.ndarray:
        """Provide default observations.

        Gives a default observations of the correct shape, purpose is when the
        observation fails that the observation can still be generated so that
        the training does not error out.

        Returns:
            np.ndarray: An array filled with ones in the correct shape and size
            of the observation.
        """
        if self.value_type == "CNN":
            return np.zeros(self.observation_shape, dtype=np.uint8)
        return (
            np.ones(self.observation_shape, dtype=np.float32) * self.value_min
        )



"""Definition of the geometry to be optimized, which informs the action space.

File holds all classes which define the geometry and with that also the action
definition of the problem.
"""

from typing import Any, List, Optional, Tuple, Union
from uuid import UUID

import numpy as np
from gymnasium import spaces
from pydantic import PrivateAttr

from releso.base_model import BaseModel
from releso.exceptions import ParserException
from releso.mesh import MeshExporter, MeshTypes
from releso.shape_parameterization import ShapeDefinition, VariableLocation
from releso.spline import BSplineDefinition, NURBSDefinition, SplineDefinition
from releso.util.types import ObservationType

try:
    from splinepy.helpme.ffd import FFD
except ImportError as err:  # pragma: no cover
    from releso.util.module_import_raiser import ModuleImportRaiser

    FFD = ModuleImportRaiser("splinepy - FFD", err)

ShapeTypes = Union[ShapeDefinition, BSplineDefinition, NURBSDefinition]


class Geometry(BaseModel):
    """Definition of the Geometry."""

    #: Definition of the shape used for the geometry.
    #: This is the shape that will be used to define the geometry and
    #: the action space. If only a single shape is used this can be
    #: a single shape definition, if multiple shapes are used this can be
    #: a list of shape definitions.
    shape_definition: Union[ShapeTypes, List[ShapeTypes]]
    #: use the action space for the observations. If this is set to False
    #: please note that you need to define your own observation via a
    #: :py:class:`releso.spor.SPORObject`
    action_based_observation: bool = True
    #: Whether or not to use discrete actions if False continuous actions
    #: will be used.
    discrete_actions: bool = True
    #: Whether or not to reset the controllable control point variables to a
    #: random state (True) or the default original state (False).
    reset_with_random_action_values: bool = False

    #: saved list of all available actions
    _actions: List[VariableLocation] = PrivateAttr()
    #: positions of all actions values of the previous step
    _last_actions: List[VariableLocation] = PrivateAttr(default=None)

    def _get_actions(self) -> List[VariableLocation]:
        """Get the actions defined by the shape definition.

        Returns:
            List[VariableLocation]: List of all actions defined by the
                shape_definition.
        """
        if isinstance(self.shape_definition, list):
            return [
                action
                for shape in self.shape_definition
                for action in shape.get_actions()
            ]
        return self.shape_definition.get_actions()

    def __init__(self, **data: Any) -> None:
        """Definition of the Geometry.

        Base geometry which just represents the used shape.

        Geometry also handles how the actions are defined and applied to the
        shape.
        """
        super().__init__(**data)
        self._actions = self._get_actions()

    def setup(self, environment_id: UUID):
        """Setup with additional information from the environment.

        Base geometry has no changes.

        Args:
            environment_id (UUID): Environment id.
        """
        self._actions = self._get_actions()

    def get_parameter_values(self) -> List[List[float]]:
        """Return all control_points of the spline.

        Returns:
            List[List[float]]: Nested list of control_points.
        """
        if isinstance(self.shape_definition, list):
            return [
                shape.get_parameter_values() for shape in self.shape_definition
            ]
        else:
            return self.shape_definition.get_parameter_values()

    def apply_action(self, action: Union[List[float], int]) -> Optional[Any]:
        """Function that applies a given action to the Spline.

        Args:
            action ([type]):  Action value depends on if the ActionSpace is
                              discrete (int - Signifier of the action) or
                              Continuous (List[float] - Value for each
                              continuous variable.)
        """
        self._last_actions = [act.current_position for act in self._actions]
        self.get_logger().debug(f"Applying action {action}")
        if self.discrete_actions:
            increasing: bool = action % 2 == 0
            action_index: int = int(action / 2)
            self.get_logger().debug(
                f"Setting discrete action, of variable {action_index}."
            )
            self._actions[action_index].apply_discrete_action(increasing)
            # TODO check if this is correct
        else:
            for new_value, action_obj in zip(action, self._actions):
                action_obj.apply_continuous_action(new_value)
        return self.apply()

    def apply(self) -> Optional[Any]:
        """Function which applies the current action values to the geometry.

        Overwrite if the geometry is represented by more than the shape
        representation.

        For the basic geometry no changes to the geometry are necessary. Since
        geometry is given by the shape directly.

        It also returns the basic information about the geometry. In this basic
        case it just returns the control_points of the shape. Do not rely on
        it returning the control_points since this might not be the case for
        subclasses. If you need the control_points use the relevant function.

        Returns:
            List[List[float]]: Current control points of the shape.
        """
        return self.get_parameter_values()

    def is_geometry_changed(self) -> bool:
        """Checks if the geometry was changed with the previous action apply.

        This function uses the actions as a metric on whether the geometry was
        changed. If the actions were not changed the function will return
        `False`, if the actions were changed it will return `True`.

        Needs to be reimplemented if other factors influence the geometry.

        Returns:
            bool: See above.
        """
        if self._last_actions is None:
            return False
        return not np.allclose(
            np.array([act.current_position for act in self._actions]),
            np.array(self._last_actions),
        )

    def get_action_definition(self) -> spaces.Space:
        """Return actions definition defined by the shape parametrization.

        Returns:
            spaces.Space: VariableLocations that have 'wriggle' room.
        """
        if self.discrete_actions:
            return spaces.Discrete(len(self._actions) * 2)
        else:
            return spaces.Box(low=-1, high=1, shape=(len(self._actions),))

    def get_observation_definition(self) -> Tuple[str, ObservationType]:
        """Return the geometry observation definition.

        The geometry observation by default just includes the action space of
        the shape definition.

        Returns:
            Tuple[str, ObservationType]: Tuple of the descriptor of the
                observation space and the gym based definition of the
                observation space.
        """
        if not self.action_based_observation:
            self.get_logger().warning(
                "Observation space is accessed which should not happen."
            )
        return "geometry_observation", spaces.Box(
            low=0, high=1, shape=(len(self._actions),), dtype=np.float32
        )

    def get_observation(self) -> Optional[np.ndarray]:
        """Returns the current observations.

        For the basic geometry it is possible to define the current values of
        actions to be the observations. This can be done via the variable
        `action_based_observation`.

        Returns:
            Optional[np.ndarray]: Optional action based observations.
        """
        if not self.action_based_observation:
            return None
        return np.array([
            var_loc.current_position for var_loc in self._actions
        ])

    def reset(self, validation_id: Optional[int] = None) -> Any:
        """Resets the geometry to its initial values."""
        if self.reset_with_random_action_values:
            self.apply_random_action(validation_id)
        else:
            if isinstance(self.shape_definition, list):
                for shape in self.shape_definition:
                    shape.reset()
            else:
                self.shape_definition.reset()
        return self.apply()

    def apply_random_action(self, seed: Optional[str] = None):
        """Apply a random continuous action.

        Applying a random continuous action to all movable control point
        variables. Can be activated to be used during the reset of an
        environment.

        Args:
            seed (Optional[str], optional): Seed for the generation of the
                random action. The same seed will result in always the same
                action. This functionality is chosen to make validation
                possible. If None (default) a random seed will be used and
                the action will be different each time. Defaults to None.
        """

        def _parse_string_to_int(string: str) -> int:
            chars_as_ints = [ord(char) for char in str(string)]
            string_as_int = sum(chars_as_ints)
            return string_as_int

        seed = _parse_string_to_int(str(seed)) if seed is not None else seed
        self.get_logger().debug(
            f"A random action is applied during reset with the following "
            f"seed {str(seed)}"
        )
        rng_gen = np.random.default_rng(seed)
        random_action = (rng_gen.random((len(self._actions),)) * 2) - 1
        for new_value, action_obj in zip(random_action, self._actions):
            action_obj.apply_continuous_action(new_value)
        return random_action


class FFDGeometry(Geometry):
    """FFD based variable shape."""

    mesh: MeshTypes
    export_mesh: Optional[MeshExporter] = None

    #: FreeFormDeformation used for the spline based shape optimization
    _FFD: None = PrivateAttr()

    def __init__(self, **data: Any) -> None:
        """Geometry based on a FFD transformation.

        This geometry uses the shape parametrization to deform the given mesh,
        via a FFD.

        Raises:
            RuntimeError: FFD can only be performed via a Spline based shape.
        """
        super().__init__(**data)

        if not issubclass(type(self.shape_definition), SplineDefinition):
            raise ParserException(
                "FFDGeometry",
                "shape_definition",
                "FFD can only be performed with a splinepy Spline",
            )
        self._FFD = FFD()

    def setup(self, environment_id: UUID):
        """Setup with additional information from the environment.

        Needs to adapt export path with the environment id.

        Args:
            environment_id (UUID): Environment id.
        """
        super().setup(environment_id=environment_id)
        self._FFD.mesh = self.mesh.get_mesh()
        if self.export_mesh:
            self.export_mesh.adapt_export_path(environment_id=environment_id)

    def apply_action(self, action: Union[List[float], int]) -> Optional[Any]:
        """Function that applies a given action to the Spline.

        Args:
            action (Union[List[float], int]):  Action value depends on if the
                ActionSpace is discrete (int - Signifier of the action) or
                Continuous (List[float] - Value for each continuous variable.)
        """
        return super().apply_action(action)

    def apply(self) -> Optional[Any]:
        """Apply the shape via FFD on to the geometry.

        Returns:
            Optional[Any]: Return the correct values.
        """
        return self.apply_ffd()

    def apply_ffd(self, path: Optional[str] = None) -> Union[str, np.ndarray]:
        """Apply FFD for the current shape.

        Might move in the future to a SPORStep. Can be deactivated with
        do_not_perform_ffd.

        Apply the Free Form Deformation using the current spline to the mesh
        and export the resulting mesh to the path given.

        Args:
            path (Optional[str], optional):
                Path to where the deformed mesh was exported to or the of no
                export wanted the vertices of the mesh. Defaults to None.

        Returns:
            Union[str, np.ndarray]: Path to the exported mesh file or the
                vertices of the mesh.
        """
        self._FFD.spline = self.shape_definition.get_shape()
        # since self._FFD.mesh (this will perform the ffd) is called in the
        # export function it does not need to be called here
        if self.export_mesh:
            self.export_mesh.export_mesh(self._FFD.mesh)
            return self.export_mesh.get_export_path()
        else:
            return self._FFD.mesh.vertices


GeometryTypes = Union[Geometry, FFDGeometry]



import json
import uuid
from pathlib import Path

import pytest
from conftest import dir_save_location_path

from releso.util.reward_helpers import (
    load_json,
    spor_com_additional_information,
    spor_com_parse_arguments,
    write_json,
)


def test_load_json(dir_save_location: Path, clean_up_provider):
    file_loc: Path = dir_save_location / "test.json"
    assert not file_loc.exists()
    dir_save_location.mkdir(parents=True, exist_ok=True)
    res = load_json(file_loc)
    assert res == {}
    assert file_loc.exists()
    test_dict = {"test": 12, "other": "test"}
    with open(file_loc, "w") as f_open:
        json.dump(test_dict, f_open)
    res = load_json(file_loc)
    assert res == test_dict
    clean_up_provider(file_loc)


def test_load_json_folder_does_not_exist(dir_save_location, clean_up_provider):
    file_loc: Path = dir_save_location / "test" / "test.json"
    with pytest.raises(RuntimeError) as err:
        load_json(file_loc)
    assert "You have to create the folder for the json first" in str(err.value)
    clean_up_provider(file_loc)


def test_write_json(dir_save_location: Path, clean_up_provider):
    dir_save_location.mkdir(parents=True, exist_ok=True)
    file_loc: Path = dir_save_location / "test.json"
    test_dict = {"test": 12, "other": "test"}
    write_json(file_loc, test_dict)

    with open(file_loc) as f_open:
        res = json.load(f_open)
    clean_up_provider(file_loc)
    assert res == test_dict


def test_additional_information():
    # the tested function removes the quotation marks around the string.
    # These appear due to it being processed by argparse
    test_dict = {"test": 12, "other": "test"}
    res = spor_com_additional_information(f"'{json.dumps(test_dict)}'")
    assert res == test_dict


@pytest.mark.parametrize(
    ["initialize", "i", "reset"],
    [
        (True, False, False),
        (False, True, False),
        (False, False, True),
    ],
)
@pytest.mark.parametrize(
    ["r", "run", "run_id", "error_r"],
    [
        (str(uuid.uuid4()), False, False, False),
        # (False, str(uuid.uuid4()), False, False),
        # (False, False, str(uuid.uuid4()), False),
        (
            False,
            False,
            False,
            "-r/--run/--run_id",
        ),
    ],
)
@pytest.mark.parametrize(
    ["v", "validation_value", "error_v"],
    [
        (
            "True",
            False,
            True,
        ),
        (
            False,
            False,
            False,
        ),
        (
            str(1.2),
            False,
            False,
        ),
        (
            False,
            "True",
            True,
        ),
    ],
)
@pytest.mark.parametrize(
    ["j", "additional_values", "json_object", "error_j"],
    [
        (
            True,
            False,
            False,
            False,
        ),
    ],
)
@pytest.mark.parametrize(
    ["l_parameter", "base_save_location", "error_l"],
    [
        (
            False,
            str(dir_save_location_path(False)),
            False,
        ),
        (
            False,
            False,
            "-l/--base_save_location",
        ),
    ],
)
@pytest.mark.parametrize(
    ["e", "environment_id", "error_e"],
    [
        (
            "test",
            False,
            False,
        ),
        (
            False,
            False,
            "-e/--environment_id",
        ),
    ],
)
def test_parse_communication_interface(
    initialize,
    i,
    reset,
    r,
    run,
    run_id,
    error_r,
    v,
    validation_value,
    error_v,
    j,
    additional_values,
    json_object,
    error_j,
    l_parameter,
    base_save_location,
    error_l,
    e,
    environment_id,
    error_e,
    dir_save_location,
    capsys,
):
    args = []
    if initialize:
        args.append("--initialize")
    elif i:
        args.append("-i")
    elif reset:
        args.append("--reset")
    if r:
        args.extend(["-r", r])
    elif run:
        args.extend(["--run", run])
    elif run_id:
        args.extend(["--run_id", run_id])
    if v:
        args.extend(["-v", v])
    elif validation_value:
        args.extend(["--validation_value", validation_value])
    if l_parameter:
        args.extend(["-l", l_parameter])
    elif base_save_location:
        args.extend(["--base_save_location", base_save_location])
    if e:
        args.extend(["-e", e])
    elif environment_id:
        args.extend(["--environment_id", environment_id])

    if any([error_r, error_v, error_j, error_l, error_e]):
        try:
            spor_com_parse_arguments(args)
        except SystemExit:
            pass
        err = capsys.readouterr().err
        errors = []
        for error in [error_r, error_v, error_j, error_l, error_e]:
            if error:
                if isinstance(error, str):
                    errors.append(error in err)
                else:
                    errors.append(True)
        assert any(errors)
        return
    argparse_ret = spor_com_parse_arguments(args)
    # raise RuntimeError(f"{argparse_ret}")
    assert argparse_ret.reset == any([initialize, i, reset])
    if run:
        assert argparse_ret.run_id == uuid.UUID(run)
    elif run_id:
        assert argparse_ret.run_id == uuid.UUID(run_id)
    elif r:
        assert argparse_ret.run_id == uuid.UUID(r)
    if validation_value:
        assert argparse_ret.validation_value == float(validation_value)
    elif v:
        assert argparse_ret.validation_value == float(v)
    if base_save_location:
        assert argparse_ret.base_save_location == base_save_location
    elif l_parameter:
        assert argparse_ret.base_save_location == l_parameter
    if environment_id:
        assert argparse_ret.environment_id == environment_id
    elif e:
        assert argparse_ret.environment_id == e


@pytest.mark.parametrize(
    ["j", "additional_values", "json_object", "error_j"],
    [
        (
            False,
            False,
            False,
            False,
        ),
        (
            False,
            f'"{json.dumps({"test": 12, "other": "test"})}"',
            False,
            False,
        ),
        (
            False,
            False,
            f'"{json.dumps({"test": 12, "other": "test"})}"',
            False,
        ),
        (
            f'"{json.dumps({"test": 12, "other": "test"})}"',
            False,
            False,
            False,
        ),
    ],
)
def test_parse_communication_interface_json(
    j, additional_values, json_object, error_j, dir_save_location, capsys
):
    args = []
    args.extend(
        [
            f"-r{str(uuid.uuid4())}",
            f"-e{str(uuid.uuid4())}",
            f"-l{str(dir_save_location)}",
        ]
    )
    if j:
        args.extend(["-j", j])
    elif additional_values:
        args.extend(["--additional_values", additional_values])
    elif json_object:
        args.extend(["--json_object", json_object])
    if error_j:
        try:
            spor_com_parse_arguments(args)
        except SystemExit:
            pass
        err = capsys.readouterr().err
        assert error_j in err
        return
    argparse_ret = spor_com_parse_arguments(args)
    if json_object:
        assert argparse_ret.json_object == {"test": 12, "other": "test"}
    elif additional_values:
        assert argparse_ret.json_object == {"test": 12, "other": "test"}
    elif j:
        assert argparse_ret.json_object == {"test": 12, "other": "test"}
    else:
        assert argparse_ret.json_object is None



"""Parsing of the agents available in ReLeSO.

Out of the box the ReLeSO package uses agents implemented in the Python
package stable-baselines3. Currently the agents Deep Q-Network (DQN), Proximal
Policy Optimization (PPO), Soft Actor-Critic (SAC) Advantage Actor Critic (A2C)
and Deep Deterministic Policy Gradient (DDPG) can be used directly but the
others can be added easily.

The following table shows which agent can be used for which shape optimization
approach:

+-------+---------------------------+--------------------------------+
| Agent | Direct shape optimization | Incremental shape optimization |
+=======+===========================+================================+
| PPO   | YES                       | YES                            |
+-------+---------------------------+--------------------------------+
| DQN   | NO                        | YES                            |
+-------+---------------------------+--------------------------------+
| SAC   | YES                       | NO                             |
+-------+---------------------------+--------------------------------+
| DDPG  | YES                       | NO                             |
+-------+---------------------------+--------------------------------+
| A2C   | Yes                       | YES                            |
+-------+---------------------------+--------------------------------+

Author:
    Clemens Fricke (clemens.david.fricke@tuwien.ac.at)

"""
import datetime
import pathlib
from typing import Any, Dict, Literal, Optional, Union

from pydantic import Field
from pydantic.types import FilePath
from stable_baselines3 import A2C, DDPG, DQN, PPO, SAC
from stable_baselines3.common.base_class import BaseAlgorithm

from releso.base_model import BaseModel
from releso.exceptions import AgentUnknownException
from releso.gym_environment import GymEnvironment
from releso.util.module_import_raiser import ModuleImportRaiser

try:
    from releso.feature_extractor import CombinedExtractor, FeatureExtractor
except ImportError as err:
    CombinedExtractor = ModuleImportRaiser("torchvision", err)
    FeatureExtractor = ModuleImportRaiser("torchvision", err)

# TODO dynamic agent detection via https://stackoverflow.com/a/3862957
####################################
# >>> def get_subclasses(agent_type):
# ...     for sub in agent_type.__subclasses__():
# ...             print(sub)
# ...             get_subclasses(sub)
# ...
# >>> get_subclasses(stable_baselines3.common.base_class.BaseAlgorithm)
# <class 'stable_baselines3.common.on_policy_algorithm.OnPolicyAlgorithm'>
# <class 'stable_baselines3.a2c.a2c.A2C'>
# <class 'stable_baselines3.ppo.ppo.PPO'>
# <class 'stable_baselines3.common.off_policy_algorithm.OffPolicyAlgorithm'>
# <class 'stable_baselines3.td3.td3.TD3'>
# <class 'stable_baselines3.ddpg.ddpg.DDPG'>
# <class 'stable_baselines3.dqn.dqn.DQN'>
####################################
# get argument names https://docs.python.org/3/library/inspect.html


class BaseAgent(BaseModel):
    """Base agent definition.

    The class BaseAgent should be used as the base class for all classes
    defining agents for the ReLeSO framework.
    """

    #: base directory of the tensorboard logs if given an experiment name
    #: with a current timestamp is also added.
    tensorboard_log: Optional[str]

    def get_next_tensorboard_experiment_name(self) -> Optional[str]:
        """Return tensorboard experiment name.

        Adds a date and time marker to the tensorboard experiment name so that
        it can be distinguished from other experiments.

        Returns:
            str: Experiment name consisting of a time and date stamp.
        """
        if self.tensorboard_log is not None:
            return "_".join(
                [
                    self.tensorboard_log,
                    datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S"),
                ]
            )
        return None


class BaseTrainingAgent(BaseAgent):
    """BaseTraining agent definition.

    The class BaseAgent should be used as the base class for all classes
    defining agents for the ReLeSO framework.
    """

    #: policy defines the network structure which the agent uses
    policy: Literal["MlpPolicy", "CnnPolicy", "MultiInputPolicy"]
    #: If given the str identifies the Custom Feature Extractor to be added.
    use_custom_feature_extractor: Optional[
        Literal[
            "resnet18",
            "mobilenet_v2",  # , "mobilenetv3_small", "mobilenetv3_large"
        ]
    ] = None
    #: use the custom feature extractor with out a final linear layer
    cfe_without_linear: bool = False
    #: additional arguments to be passed to the policy on creation
    policy_kwargs: Optional[Dict[str, Any]] = None

    def get_additional_kwargs(self, **kwargs) -> Dict[str, Any]:
        """Add additional keyword arguments for agent instantiation.

        Reads and gets the additional keyword arguments for the agent
        definition.

        Returns:
            Dict[str, Any]: Dictionary of the needed additional keywords.
        """
        if self.policy_kwargs is None:
            self.policy_kwargs = {}
        if self.use_custom_feature_extractor:
            self.policy_kwargs["features_extractor_kwargs"] = {}
            if self.policy == "CnnPolicy":
                self.policy_kwargs[
                    "features_extractor_class"
                ] = FeatureExtractor
            elif self.policy == "MultiInputPolicy":
                self.policy_kwargs[
                    "features_extractor_class"
                ] = CombinedExtractor
            else:
                self.get_logger().warning(
                    "Please use the CnnPolicy or the MultiInputPolicy with "
                    "the FeatureExtractors. The MlpPolicy - if it works - will"
                    " use fully connected layers which might produce large "
                    "networks with image data as input."
                )
                self.policy_kwargs[
                    "features_extractor_class"
                ] = FeatureExtractor
            if self.cfe_without_linear:
                self.policy_kwargs["features_extractor_kwargs"][
                    "without_linear"
                ] = True
            self.policy_kwargs["features_extractor_kwargs"][
                "network_type"
            ] = self.use_custom_feature_extractor
            self.policy_kwargs["features_extractor_kwargs"][
                "logger"
            ] = self.get_logger()
        # add all object variables to the kwargs, which are not in the
        # blacklist
        return_dict = {
            k: v
            for k, v in self.__dict__.items()
            if k
            not in [
                "agent_type",
                "logger_name",
                "save_location",
                "use_custom_feature_extractor",
                "cfe_without_linear",
                "save_location",
            ]
        }
        return return_dict


class PretrainedAgent(BaseAgent):
    """Pretrained agent definition.

    This class can be used to load pretrained agents, instead of using
    untrained agents. Can also be used to only validate this agent without
    training it further. Please see validation section for this use-case.
    """

    #: What RL algorithm was used to train the agent. Needs to be know to
    #: correctly load the agent.
    agent_type: Literal["PPO", "SAC", "DDPG", "A2C", "DQN"] = Field(
        alias="type"
    )
    #: Path to the save files of the pretrained agent.
    path: Union[FilePath, pathlib.Path]
    #: If the agent is to be trained further the results can be added to the
    # existing tensorboard experiment. This is the path to the existing
    # tensorboard experiment
    tesorboard_run_directory: Union[str, pathlib.Path, None] = None

    # TODO add testing of agent loading. Need trained agents?
    def get_agent(
        self, environment: GymEnvironment, normalizer_divisor: int = 1
    ) -> BaseAlgorithm:  # pragma: no cover
        """Tries to locate the agent defined and to load it correctly.

        Args:
            environment (GymEnvironment): Environment with which the agent will
            interact.
            normalizer_divisor (int): Currently not used in this function.

        Raises:
            AgentUnknownException: Possible exception which can be
            thrown.

        Returns:
            BaseAlgorithm: Return the correctly loaded agent.
        """
        self.get_logger().info(
            f"Using pretrained agent of type {self.agent_type} from location "
            f"{self.path}."
        )
        if self.agent_type == "PPO":
            return PPO.load(self.path, environment)
        elif self.agent_type == "DDPG":
            return DDPG.load(self.path, environment)
        elif self.agent_type == "SAC":
            return SAC.load(self.path, environment)
        elif self.agent_type == "DQN":
            return DQN.load(self.path, environment)
        elif self.agent_type == "A2C":
            return A2C.load(self.path, environment)
        else:
            raise AgentUnknownException(self.agent_type)

    def get_next_tensorboard_experiment_name(self) -> Optional[str]:
        """Return the name of the tensorboard experiment.

        The tensorboard experiment name of the original training run if given
        else a new one with the current time stamp.

        Returns:
            str: tensorboard experiment name
        """
        if self.tesorboard_run_directory:
            return str(self.tesorboard_run_directory) if not None else None
        if self.tensorboard_log is not None:
            return super().get_next_tensorboard_experiment_name()
        return None


class A2CAgent(BaseTrainingAgent):
    """A2c agent definition.

    A2C definition for the stable_baselines3 implementation for this algorithm.
    Variable comments are taken from the stable_baselines3 documentation.
    """

    #: What RL algorithm was used to train the agent. Needs to be know to
    #: correctly load the agent.
    agent_type: Literal["A2C"] = Field(alias="type")
    #: The learning rate, it can be a function of the current progress
    #: remaining (from 1 to 0)
    learning_rate: float = 7e-4
    #: The number of steps to run for each environment per update(i.e. rollout
    #: buffer size is n_steps * n_envs where n_envs is number of environment
    #: copies running in parallel) NOTE: n_steps * n_envs must be greater than
    #: 1 (because of the advantage normalization)
    #: See https://github.com/pytorch/pytorch/issues/29372
    n_steps: int = 5
    #: Discount factor
    gamma: float = 0.99
    #: Factor for trade-off of bias vs variance for Generalized Advantage
    #: Estimator to classic advantage when set to 1
    gae_lambda: float = 1.0
    #: Entropy coefficient for the loss calculation
    ent_coef: float = 0.0
    #: Value function coefficient for the loss calculation
    vf_coef: float = 0.5
    #: The maximum value fot he gradient clipping
    max_grad_norm = 0.5
    #: RMSProp epsilon. It stabilizes square root computation in denominator
    #: of of RMSProp update
    rms_prop_eps = 1e-05
    #: Whether to use RMSprop (default) or Adam as optimizer
    use_rms_prop = True
    #: Whether to use generalized State Dependent Exploration (gSDE) instead
    #: of action noise exploration (default: False)
    use_sde = False
    #: Sample a new noise matrix every n steps when using gSDE Default: -1
    #: (sample only at beginning of roll)
    sde_sample_freq = -1
    #: Whether to normalize or not the advantage
    normalize_advantage = False
    #: Seed for the pseudo random generators
    seed: Optional[int] = None
    #: Device (cpu, cuda, …) on which the code should be run. Setting it to
    #: auto, the code will be run on the GPU if possible.
    device: str = "auto"

    def get_agent(
        self, environment: GymEnvironment, normalizer_divisor: int = 1
    ) -> A2C:
        """Creates the stable_baselines version of the wanted Agent.

        Uses all Variables given in the object (except agent_type) as the input
        parameters of the agent object creation.

        Notes:
            The A2C variable n_steps scales with the amount of environments the
            agent is trained with. If this behavior is unwanted you can set the
            variable ''normalize_training_values'' in BaseParser to true. This
            will set this variable to the number of environments so that the
            scaled values can be unscaled.

        Args:
            environment (GymEnvironment):
                The environment the agent uses to train.
            normalizer_divisor (int):
                Divides the variable n_steps by this value to descale scaled
                values with n_environments unequal zero.

        Returns:
            A2C: Initialized A2C agent.
        """
        self.get_logger().info(f"Using agent of type {self.agent_type}.")
        if normalizer_divisor == 0:
            self.get_logger().warning("Normalizer divisor is 0, will use 1.")
            normalizer_divisor = 1
        self.n_steps = max(int(self.n_steps / normalizer_divisor), 1)
        return A2C(env=environment, **self.get_additional_kwargs())


class PPOAgent(BaseTrainingAgent):
    """PPO agent definition.

    PPO definition for the stable_baselines3 implementation for this algorithm.
    Variable comments are taken from the stable_baselines3 documentation.
    """

    #: What RL algorithm was used to train the agent. Needs to be know to
    #: correctly load the agent.
    agent_type: Literal["PPO"] = Field(alias="type")
    #: The learning rate, it can be a function of the current progress
    #: remaining (from 1 to 0)
    learning_rate: float = 3e-4
    #: The number of steps to run for each environment per update(i.e.
    #: rollout buffer size is n_steps * n_envs where n_envs is number of
    #: environment copies running in parallel) NOTE: n_steps * n_envs must be
    #: greater than 1 (because of the advantage normalization) See
    #: https://github.com/pytorch/pytorch/issues/29372
    n_steps: int = 2048
    #: Minibatch size
    batch_size: Optional[int] = 64
    #: Number of epoch when optimizing the surrogate loss
    n_epochs: int = 10
    #: Discount factor
    gamma: float = 0.99
    #: Factor for trade-off of bias vs variance for Generalized Advantage
    #: Estimator
    gae_lambda: float = 0.95
    #: Clipping parameter, it can be a function of the current progress
    #: remaining (from 1 to 0).
    clip_range: float = 0.2
    #: Entropy coefficient for the loss calculation
    ent_coef: float = 0.0
    #: Value function coefficient for the loss calculation
    vf_coef: float = 0.5
    #: Seed for the pseudo random generators
    seed: Optional[int] = None
    #: Device (cpu, cuda, …) on which the code should be run. Setting it to
    #: auto, the code will be run on the GPU if possible.
    device: str = "auto"

    def get_agent(
        self, environment: GymEnvironment, normalizer_divisor: int = 1
    ) -> PPO:
        """Creates the stable_baselines version of the wanted Agent.

        Uses all Variables given in the object (except agent_type) as the input
        parameters of the agent object creation.

        Notes:
            The PPO variable n_steps scales with the amount of environments the
            agent is trained with. If this behavior is unwanted you can set the
            variable ''normalize_training_values'' in BaseParser to true. This
            will set this variable to the number of environments so that the
            scaled values can be unscaled.

        Args:
            environment (GymEnvironment):
                The environment the agent uses to train.
            normalizer_divisor (int):
                Divides the variable n_steps by this value to descale scaled
                values with n_environments unequal zero.

        Returns:
            PPO: Initialized PPO agent.
        """
        self.get_logger().info(f"Using agent of type {self.agent_type}.")
        if normalizer_divisor == 0:
            self.get_logger().warning("Normalizer divisor is 0, will use 1.")
            normalizer_divisor = 1
        self.n_steps = max(int(self.n_steps / normalizer_divisor), 1)
        return PPO(env=environment, **self.get_additional_kwargs())


class DDPGAgent(BaseTrainingAgent):
    """DDPG agent definition.

    DDPG definition for the stable_baselines3 implementation for this
    algorithm. Variable comments are taken from the stable_baselines3
    documentation.
    """

    #: What RL algorithm was used to train the agent. Needs to be know to
    #: correctly load the agent.
    agent_type: Literal["DDPG"] = Field(alias="type")
    #: The learning rate, it can be a function of the current progress
    #: remaining (from 1 to 0)
    learning_rate: float = 1e-3
    #: size of the replay buffer
    buffer_size: int = 100000
    #: how many steps of the model to collect transitions for before learning
    #: starts
    learning_starts: int = 100
    #: Minibatch size
    batch_size: Optional[int] = 64
    #: the soft update coefficient ("Polyak update", between 0 and 1)
    tau: float = 0.005
    #: Discount factor
    gamma: float = 0.99
    #: Enable a memory efficient variant of the replay buffer at a cost of more
    #: complexity. See
    #: https://github.com/DLR-RM/stable-baselines3/issues/37#:issuecomment
    # -637501195
    optimize_memory_usage: bool = False
    #: Seed for the pseudo random generators
    seed: Optional[int] = None
    #: Device (cpu, cuda, …) on which the code should be run. Setting it to
    #: auto, the code will be run on the GPU if possible.
    device: str = "auto"

    def get_agent(
        self, environment: GymEnvironment, normalizer_divisor: int = 1
    ) -> DDPG:
        """Creates the stable_baselines version of the wanted Agent.

        Uses all Variables given in the object (except agent_type) as the input
        parameters of the agent object creation.

        Args:
            environment (GymEnvironment): The environment the agent uses to
            train.
            normalizer_divisor (int): Currently not used in this function.

        Returns:
            DDPG: Initialized DDPG agent.
        """
        self.get_logger().info(f"Using agent of type {self.agent_type}.")
        return DDPG(env=environment, **self.get_additional_kwargs())


class SACAgent(BaseTrainingAgent):
    """SAC Agent definition.

    SAC definition for the stable_baselines3 implementation for this algorithm.
    Variable comments are taken from the stable_baselines3 documentation.
    """

    #: What RL algorithm was used to train the agent. Needs to be know to
    #: correctly load the agent.
    agent_type: Literal["SAC"] = Field(alias="type")
    #: The learning rate, it can be a function of the current progress
    #: remaining (from 1 to 0)
    learning_rate: float = 1e-3
    #: size of the replay buffer
    buffer_size: int = 1000000
    #: how many steps of the model to collect transitions for before learning
    #: starts
    learning_starts: int = 100
    #: Minibatch size
    batch_size: Optional[int] = 64
    #: the soft update coefficient ("Polyak update", between 0 and 1)
    tau: float = 0.005
    #: Discount factor
    gamma: float = 0.99
    #: Enable a memory efficient variant of the replay buffer at a cost of more
    #: complexity. See https://github.com/DLR-RM/stable-baselines3/issues/37#:
    # issue comment-637501195
    optimize_memory_usage: bool = False
    #: Entropy regularization coefficient. (Equivalent to inverse of reward
    #: scale in the original SAC paper.)  Controlling exploration/exploitation
    #: trade-off. Set it to 'auto' to learn it automatically (and 'auto_0.1'
    #: for using 0.1 as initial value)
    ent_coef: Union[str, float] = "auto"
    #: update the target network every ``target_network_update_freq`` gradient
    #: steps.
    target_update_interval: int = 1
    #: target entropy when learning ``ent_coef`` (``ent_coef = 'auto'``)
    target_entropy: Union[str, float] = "auto"
    #: Whether to use generalized State Dependent Exploration (gSDE) instead
    #: of action noise exploration (default: False)
    use_sde: bool = False
    #: Sample a new noise matrix every n steps when using gSDE Default: -1
    #: (only sample at the beginning of the rollout)
    sde_sample_freq: int = -1
    #: Whether to use gSDE instead of uniform sampling during the warm up
    #: phase (before learning starts)
    use_sde_at_warmup: bool = False
    #: Seed for the pseudo random generators
    seed: Optional[int] = None
    #: Device (cpu, cuda, …) on which the code should be run. Setting it to
    #: auto, the code will be run on the GPU if possible.
    device: str = "auto"

    def get_agent(
        self, environment: GymEnvironment, normalizer_divisor: int = 1
    ) -> SAC:
        """Creates the stable_baselines version of the wanted Agent.

        Uses all Variables given in the object (except agent_type) as the input
        parameters of the agent object creation.

        Args:
            environment (GymEnvironment): The environment the agent uses to
            train.
            normalizer_divisor (int): Currently not used in this function.

        Returns:
            SAC: Initialized SAC agent.
        """
        self.get_logger().info(f"Using agent of type {self.agent_type}.")
        return SAC(env=environment, **self.get_additional_kwargs())


class DQNAgent(BaseTrainingAgent):
    """DQN Agent definition.

    DQN definition for the stable_baselines3 implementation for this algorithm.
    Variable comments are taken from the stable_baselines3 documentation.
    """

    #: What RL algorithm was used to train the agent. Needs to be know to
    #: correctly load the agent.
    agent_type: Literal["DQN"] = Field(alias="type")
    #: The learning rate, it can be a function of the current progress
    #: remaining (from 1 to 0)
    learning_rate: float = 1e-4
    #: size of the replay buffer
    buffer_size: int = 1000000
    #: how many steps of the model to collect transitions for before learning
    #: starts
    learning_starts: int = 256
    #: Minibatch size for each gradient update
    batch_size: Optional[int] = 32
    #: the soft update coefficient ("Polyak update", between 0 and 1) default
    #: 1 for hard update
    tau: float = 1.0
    #: the discount factor
    gamma: float = 0.99
    #: Update the model every ``train_freq`` steps.
    train_freq: int = 4
    #: How many gradient steps to do after each rollout (see ``train_freq``)
    #: Set to ``-1`` means to do as many gradient steps as steps done in the
    #: environment during the rollout.
    gradient_steps: int = 1
    #: Enable a memory efficient variant of the replay buffer at a cost of more
    #: complexity. See https://github.com/DLR-RM/stable-baselines3/issues/37#:
    # issue comment-637501195
    optimize_memory_usage: bool = False
    #: update the target network every ``target_update_interval``
    target_update_interval: int = 256
    #: fraction of entire training period over which the exploration rate is
    #: reduced
    exploration_fraction: float = 0.1
    #: initial value of random action probability
    exploration_initial_eps: float = 1.0
    #: final value of random action probability
    exploration_final_eps: float = 0.05
    #: The maximum value for the gradient clipping
    max_grad_norm: float = 10
    #: Seed for the pseudo random generators
    seed: Optional[int] = None
    #: Device (cpu, cuda, …) on which the code should be run. Setting it to
    #: auto, the code will be run on the GPU if possible.
    device: str = "auto"

    def get_agent(
        self, environment: GymEnvironment, normalizer_divisor: int = 1
    ) -> DQN:
        """Creates the stable_baselines version of the wanted Agent.

        Uses all Variables given in the object (except agent_type) as the input
        parameters of the agent object creation.

        Args:
            environment (GymEnvironment): The environment the agent uses to
            train.
            normalizer_divisor (int): Currently not used in this function.

        Returns:
            DQN: Initialized DQN agent.
        """
        self.get_logger().info(f"Using agent of type {self.agent_type}.")
        return DQN(env=environment, **self.get_additional_kwargs())


AgentTypeDefinition = Union[
    PPOAgent, DDPGAgent, SACAgent, PretrainedAgent, DQNAgent, A2CAgent
]



from collections import OrderedDict

import numpy as np
import pydantic
import pytest
from conftest import dir_save_location_path

from releso.shape_parameterization import ShapeDefinition, VariableLocation
from releso.util.logger import VerbosityLevel


@pytest.mark.parametrize(
    "current_position, error",
    [
        (0, False),
        (100, False),
        (-100, False),
        (0.001, False),
        (-0.001, False),
        (None, True),  # Need at least current_position to instantiate
    ],
)
def test_variable_location_instantiates_only_current_position(
    current_position, error, dir_save_location
):
    """Test that VariableLocation instantiates"""
    # First check the error state.
    cal_dict = {
        "current_position": current_position,
        "save_location": dir_save_location,
    }
    if error:
        with pytest.raises(pydantic.ValidationError):
            variable_location = VariableLocation(**cal_dict)
        return
    # check default initiation
    variable_location = VariableLocation(**cal_dict)
    assert variable_location is not None
    assert variable_location.current_position == current_position
    assert variable_location.is_action() is False
    assert variable_location.min_value == current_position
    assert variable_location.max_value == current_position
    assert variable_location.n_steps == 10
    assert current_position == variable_location._original_position


@pytest.mark.parametrize(
    "current_position, min_value, max_value, error",
    [
        (0, None, None, False),
        (0, -10, None, False),
        (-10, -12, -8, False),
        (-10, -8, None, True),
        (-10, -8, -7, True),
        (-10, None, -12, True),
    ],
)
def test_variable_location_min_max_location(
    current_position, min_value, max_value, error, dir_save_location
):
    """Test for correct setting of the min and max values."""
    # check if error is emitted
    cal_dict = {
        "current_position": current_position,
        "min_value": min_value,
        "max_value": max_value,
        "save_location": dir_save_location,
    }
    if error:
        with pytest.raises(pydantic.ValidationError) as exc_info:
            variable_location = VariableLocation(**cal_dict)
        assert "or equal to the" in str(exc_info.value)
        return
    variable_location = VariableLocation(**cal_dict)
    if min_value is None:
        assert current_position == variable_location.min_value
    else:
        assert min_value == variable_location.min_value
    if max_value is None:
        assert current_position == variable_location.max_value
    else:
        assert max_value == variable_location.max_value


@pytest.mark.parametrize(
    "current_position, min_value, max_value, is_action",
    [
        (0, -1, 1, True),
        (0, 0, 1, True),
        (0, None, None, False),
        (0, -1, 0, True),
    ],
)
def test_variable_location_is_action(
    current_position, min_value, max_value, is_action, dir_save_location
):
    variable_location = VariableLocation(
        **{
            "current_position": current_position,
            "min_value": min_value,
            "max_value": max_value,
            "save_location": dir_save_location,
        }
    )
    assert variable_location.is_action() is is_action


@pytest.mark.parametrize(
    "current_position, min_value, max_value, n_steps, step, expected_step, "
    "error",
    [
        (0, -1, 1, None, None, 0.2, False),
        (0, -1, 1, None, 0.1, 0.1, False),
        (0, -2, 2, 100, None, 4 / 100, False),
        (0, -2, 2, 100, 0.25, 0.25, False),
    ],
)
def test_variable_location_step_length(
    current_position,
    min_value,
    max_value,
    n_steps,
    step,
    expected_step,
    error,
    dir_save_location,
):
    """Test for correct setting of the n_steps and step values."""
    # check if error is emitted
    cal_dict = {
        "current_position": current_position,
        "min_value": min_value,
        "max_value": max_value,
        "n_steps": n_steps,
        "step": step,
        "save_location": dir_save_location,
    }
    if error:
        with pytest.raises(pydantic.ValidationError) as exc_info:
            variable_location = VariableLocation(**cal_dict)
        assert "or equal to the" in str(exc_info.value)
        return
    variable_location = VariableLocation(**cal_dict)
    assert np.isclose(expected_step, variable_location.step)
    if n_steps is None and step is None:
        assert n_steps == variable_location.n_steps


@pytest.mark.parametrize(
    "current_position, min_value, max_value, step, warning",
    [
        (0, -4, 1, 0.2, False),
        (0, -1, 1, 4, True),  # step length is greater than the interval
        (0, -2, 2, 4, False),
        (0, -2, 2, 4.0001, True),  # step length is greater than the interval
    ],
)
def test_variable_location_step_length_warning(
    current_position,
    min_value,
    max_value,
    step,
    warning,
    caplog,
    dir_save_location,
):
    cal_dict = {
        "current_position": current_position,
        "min_value": min_value,
        "max_value": max_value,
        "step": step,
        "save_location": dir_save_location,
    }
    if warning:
        with caplog.at_level(VerbosityLevel.WARNING, logger="ReLeSO_parser"):
            VariableLocation(**cal_dict)
            assert "is greater than the interval for" in caplog.text
    else:
        VariableLocation(**cal_dict)


@pytest.mark.parametrize(
    "current_position, min_value, max_value, n_steps, step, iterations, "
    "expect_position",
    [
        (
            0,
            None,
            None,
            None,
            None,
            [True] * 5,
            0,
        ),  # stay in bound upper limit
        (
            0,
            None,
            None,
            None,
            None,
            [False] * 5,
            0,
        ),  # stay in bound lower limit
        (0, -10, None, None, None, [True] * 5, 0),  # stay in bound upper limit
        (
            0,
            -10,
            None,
            None,
            None,
            [False] * 5,
            -5,
        ),  # correct value after iteration low
        (
            0,
            None,
            10,
            None,
            None,
            [True] * 5,
            5,
        ),  # correct value after iteration up
        (0, None, 10, None, None, [False] * 5, 0),  # lower bound w boundary
        (0, -10, None, None, None, [True] * 15, 0),  # upper bound w boundary
        (
            0,
            -10,
            None,
            None,
            None,
            [False] * 15,
            -10,
        ),  # lower bound w boundary
        (0, None, 10, None, None, [True] * 15, 10),  # upper bound w boundary
        (0, None, 10, None, None, [False] * 15, 0),  # lower bound w boundary
        (0, -1, 4, None, None, [True] * 3, 1.5),  # complex case
        (
            0,
            -1,
            4,
            None,
            None,
            [False, False, False, True, True, True, True, True],
            1.5,
        ),  # complex case back and forth
        (
            0,
            -1,
            4,
            None,
            None,
            [
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                True,
                False,
            ],
            3.5,
        ),  # complex case back and forth
    ],
)
def test_variable_location_discrete_action(
    current_position,
    min_value,
    max_value,
    n_steps,
    step,
    iterations,
    expect_position,
    dir_save_location,
):
    variable_location = VariableLocation(
        **{
            "current_position": current_position,
            "min_value": min_value,
            "max_value": max_value,
            "n_steps": n_steps,
            "step": step,
            "save_location": dir_save_location,
        }
    )
    last_current_position = current_position
    for iteration in iterations:
        last_current_position = variable_location.apply_discrete_action(
            iteration
        )
    assert np.isclose(
        last_current_position, variable_location.current_position
    )
    assert np.isclose(expect_position, variable_location.current_position)
    # check if original position is correctly preserved
    assert current_position == variable_location._original_position
    # check if reset works as expected
    variable_location.reset()
    assert variable_location._original_position == current_position
    assert variable_location.current_position == current_position


@pytest.mark.parametrize(
    "current_position, min_value, max_value, n_steps, step, iterations, "
    "expect_position",
    [
        # iterations are values in range [-1,1] defining the value_range of
        # the variable_location
        (10, 5, 15, None, None, [1], 15),  # upper limit
        (10, 5, 15, None, None, [0], 10),  # middle, start from middle
        (10, 5, 15, None, None, [-1], 5),  # lower limit
        (0, None, None, None, None, [1.1, -1.1], 0),  # stay in bound
        (0, -10, None, None, None, [0, 2], 0),  # stay in bound upper
        (
            0,
            -10,
            None,
            None,
            None,
            [2, 0],
            -5,
        ),  # two step correct value exceed upper
        (
            0,
            None,
            10,
            None,
            None,
            [-2, 0],
            5,
        ),  # two step correct value exceed lower
        (0, None, 10, None, None, [0.8, -1.4], 0),  # two step lower limit
        (0, -10, None, None, None, [-1.8, 1], 0),  # two step upper limit
        (0, -1, 4, None, None, [-0.33, -0.467999999], 0.33),  # complex value
    ],
)
def test_variable_location_continuous_action(
    current_position,
    min_value,
    max_value,
    n_steps,
    step,
    iterations,
    expect_position,
    dir_save_location,
):
    variable_location = VariableLocation(
        **{
            "current_position": current_position,
            "min_value": min_value,
            "max_value": max_value,
            "n_steps": n_steps,
            "step": step,
            "save_location": dir_save_location,
        }
    )
    last_current_position = current_position
    for iteration in iterations:
        last_current_position = variable_location.apply_continuous_action(
            iteration
        )
    assert np.isclose(
        last_current_position, variable_location.current_position
    )
    assert np.isclose(expect_position, variable_location.current_position)
    # check if original position is correctly preserved
    assert current_position == variable_location._original_position
    # check if reset works as expected
    variable_location.reset()
    assert variable_location._original_position == current_position
    assert variable_location.current_position == current_position


# test ShapeDefinition


@pytest.mark.parametrize(
    "control_points, n_elements, error",
    [
        ([[0.0], [1.0], [1.0]], 3, False),
        (
            [[0.0, 0.1, 0.2, 0.3, 0.4], [1.0], [1.0]],
            15,
            False,
        ),  # this might change
        ([[1.0]], 1, False),
        ([[(1.0, 2.0)]], 1, True),
        (
            [
                [
                    VariableLocation(
                        current_position=1,
                        save_location=dir_save_location_path(),
                    )
                ]
            ],
            1,
            False,
        ),
        (
            [
                [
                    VariableLocation(
                        current_position=1,
                        save_location=dir_save_location_path(),
                    ),
                    1.0,
                ]
            ],
            2,
            False,
        ),
        (
            [
                [
                    {
                        "current_position": 1,
                        "save_location": dir_save_location_path(),
                    }
                ]
            ],
            1,
            False,
        ),
        (
            [
                [
                    OrderedDict(
                        {
                            "current_position": 1,
                            "save_location": dir_save_location_path(),
                        }
                    )
                ]
            ],
            1,
            False,
        ),
    ],
)
def test_shape_definition_initiation(
    control_points, n_elements, error, dir_save_location
):
    # for idx in range(len(control_points)):
    #     for idy in range(len(control_points[idx])):
    #         if isinstance(control_points[idx][idy], dict):
    #             control_points[idx][idy]["save_location"] = dir_save_location
    #             control_points[idx][idy] = VariableLocation(
    #                 **control_points[idx][idy]
    #             )
    cal_dict = {
        "control_points": control_points,
        "save_location": dir_save_location,
    }
    if error:
        with pytest.raises(pydantic.ValidationError) as err:
            shape = ShapeDefinition(**cal_dict)
            assert "either a float or a VariableLocation" in str(err.value)
        return
    shape = ShapeDefinition(**cal_dict)
    all_variables_variable_location = True
    for element in shape.control_points:
        for item in element:
            if not isinstance(item, VariableLocation):
                all_variables_variable_location = False
                break
        if not all_variables_variable_location:
            break
    assert all_variables_variable_location

    assert shape.get_number_of_points() == n_elements

    cps = shape.get_parameter_values()
    all_the_same = True
    for original_cp_list, cp_list in zip(control_points, cps):
        for point_o, point in zip(original_cp_list, cp_list):
            if not (
                (
                    isinstance(point_o, VariableLocation)
                    and point_o.current_position == point
                )
                or (
                    isinstance(point_o, (dict, OrderedDict))
                    and point_o["current_position"] == point
                )
                or point_o == point
            ):
                all_the_same = False
                break
        if not all_the_same:
            break
    assert all_the_same

    assert shape.get_shape() == shape.get_parameter_values()


@pytest.mark.parametrize(
    "control_points, n_actions",
    [
        (
            [
                [
                    {
                        "current_position": 0.1,
                        "min_value": 0,
                        "save_location": "test",
                    },
                    {
                        "current_position": 0.2,
                        "min_value": 0,
                        "save_location": "test",
                    },
                    {
                        "current_position": 0.3,
                        "min_value": 0,
                        "save_location": "test",
                    },
                    {
                        "current_position": 0.4,
                        "min_value": 0,
                        "save_location": "test",
                    },
                    {
                        "current_position": 0.5,
                        "min_value": -1,
                        "save_location": "test",
                    },
                ]
            ],
            5,
        ),
        (
            [
                [
                    {"current_position": 0.1, "save_location": "test"},
                    {"current_position": 0.2, "save_location": "test"},
                    {
                        "current_position": 0.3,
                        "min_value": 0,
                        "save_location": "test",
                    },
                    {
                        "current_position": 0.4,
                        "min_value": 0,
                        "save_location": "test",
                    },
                    {
                        "current_position": 0.5,
                        "min_value": -1,
                        "save_location": "test",
                    },
                ]
            ],
            3,
        ),
    ],
)
def test_shape_definition_get_actions(
    control_points, n_actions, dir_save_location
):
    new_control_points = []
    for points in control_points:
        new_cps = []
        for point in points:
            point["save_location"] = dir_save_location
            new_cps.append(VariableLocation(**point))
        new_control_points.append(new_cps)
    shape = ShapeDefinition(
        **{
            "control_points": new_control_points,
            "save_location": dir_save_location,
        }
    )
    actions = shape.get_actions()
    assert len(actions) == n_actions

    original_cps = shape.get_parameter_values()

    for action in actions:
        action.apply_continuous_action(0.2)

    for o_cp, cp in zip(original_cps, shape.get_parameter_values()):
        assert o_cp != cp

    shape.reset()

    for o_cp, cp in zip(original_cps, shape.get_parameter_values()):
        assert o_cp == cp



import pandas as pd
import pytest
from stable_baselines3 import PPO

from releso.callback import EpisodeLogCallback, StepLogCallback


def test_callback_episode_log_callback(
    dir_save_location,
    clean_up_provider,
    get_null_logger,
    provide_dummy_environment,
):
    # this test is not very good, but it is a start
    # TODO: improve this test
    call_back = EpisodeLogCallback(
        episode_log_location=dir_save_location / "test.csv",
        logger=get_null_logger,
    )
    assert call_back.episode_log_location == dir_save_location / "test.csv"
    assert call_back.episodes == -1
    assert call_back.update_n_episodes == 1
    env = provide_dummy_environment
    agent = PPO("MlpPolicy", env, verbose=0, n_steps=100)
    agent.learn(100, callback=call_back)
    episode_log = pd.read_csv(dir_save_location / "test.csv")
    assert episode_log.shape[0] == 10
    clean_up_provider(dir_save_location)


@pytest.mark.parametrize(
    [
        "update_n_steps",
    ],
    [
        (0,),
        (20,),
    ],
)
def test_callback_step_information_log_callback(
    dir_save_location,
    clean_up_provider,
    get_null_logger,
    provide_dummy_environment,
    update_n_steps,
):
    # this test is not very good, but it is a start
    # TODO: improve this test
    call_back = StepLogCallback(
        step_log_location=dir_save_location / "test.jsonl",
        update_n_steps=update_n_steps,
        log_infos=True,
        logger=get_null_logger,
    )
    assert call_back.step_log_location == dir_save_location / "test.jsonl"
    assert call_back.current_episodes == []
    assert call_back.update_n_episodes == update_n_steps
    assert call_back.first_export
    env = provide_dummy_environment
    agent = PPO("MlpPolicy", env, verbose=0, n_steps=100)
    agent.learn(100, callback=call_back)
    episode_log = pd.read_json(
        dir_save_location / "test.jsonl",
        lines=True,
    )
    assert episode_log.shape[0] == 100
    clean_up_provider(dir_save_location)



"""This is the ReLeSO package. This is the best!

Please view the Documentation under :doc:`/index` for the information
about this library and framework.
"""
from releso.__version__ import __version__ as __version__



"""File hold definition for verbosity settings.

File holding the class defining the verbosity of the problem and defining the
loggers.
"""
import datetime
import pathlib
from typing import Any, Literal

from pydantic import validator
from pydantic.fields import PrivateAttr

from releso.base_model import BaseModel
from releso.util.logger import VerbosityLevel, logging, set_up_logger


class Verbosity(BaseModel):
    """Verbosity class.

    Defines the settings for the different loggers used in the current
    experiment. This class is the only class which is copied to all
    children. (this happens outside of the the standard channels and will
    hopefully not break with multiprocessing)

    Please note, the parser logger only ever can have the following name
    ``ReLeSO_parser``
    """

    #: VerbosityLevel of the parser logger. This logger should only generate
    #: messages during the setup of the experiment.
    parser: Literal["ERROR", "WARNING", "DEBUG", "INFO"] = "INFO"
    #: VerbosityLevel of the environment parsers. These loggers will generate
    #: messages during the execution of the experiments.
    #: (Training and Validation)
    environment: Literal["ERROR", "WARNING", "DEBUG", "INFO"] = "INFO"
    #: Path where the log files should be saved to. Will be inside the
    #: base_save_location.
    logfile_location: str = "logging/"
    #: Whether or not to also print the log messages to the console.
    console_logging: bool = False
    #: Base name of all logger. Defaults to "ReLeSO"
    base_logger_name: str = "ReLeSO"
    #: Name extensions for all environment logger. Defaults to "environment"
    environment_extension: str = "environment"

    # private fields
    #: save different loggers for the different instances
    _environment_logger: str = PrivateAttr(default="")
    #: save different loggers for the different instances
    _environment_validation_logger: str = PrivateAttr(default="")

    @validator("parser", "environment", always=True)
    @classmethod
    def convert_literal_str_to_verbosity_level(cls, v):
        """Validator for parser and environment variable.

        Validation function converting the string representation of the enum
        to the correct enum item.

        Args:
            v (str)): String representation of the VerbosityLevel

        Returns:
            VerbosityLevel: Enum object of the wanted value
        """
        if v == "ERROR":
            return VerbosityLevel.ERROR
        elif v == "WARNING":
            return VerbosityLevel.WARNING
        elif v == "DEBUG":
            return VerbosityLevel.DEBUG
        elif v == "INFO":
            return VerbosityLevel.INFO

    @validator("logfile_location", always=True)
    @classmethod
    def make_logfile_location_absolute(cls, v, values):
        """Validation function resolves and makes the log path absolute.

        Also adds the current timestamp to the log path if a {} is present in
        the given path.

        Args:
            v (str): [description]
            values ([type]): [description]

        Returns:
            pathlib.Path:
                pathlib representation of the path with if applicable the
                current timestamp
        """
        path: pathlib.Path = None
        v = v.format(datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S"))
        if "save_location" in values:
            # the path where the logger should write to is the
            # save_location/logfile_location
            path = values["save_location"] / v
        else:
            # the path where the logger should write to is the
            # calling_folder/logfile_location
            path = pathlib.Path(v).expanduser().resolve()
        path.mkdir(parents=True, exist_ok=True)
        return path

    def __init__(self, **data: Any) -> None:
        """Constructor verbosity parser."""
        super().__init__(**data)
        # create parser logger
        parser_name = "releso_parser"

        parser_logger = set_up_logger(
            parser_name,
            self.logfile_location,
            self.parser,
            self.console_logging,
        )

        # create base rl logger
        self._environment_logger = "_".join(
            filter(
                ("").__ne__,
                [self.base_logger_name, self.environment_extension],
            )
        )
        self.add_environment_logger_with_name_extension("")
        # create base rl logger for validation environment
        self._environment_validation_logger = "_".join(
            filter(
                ("").__ne__,
                [
                    self.base_logger_name,
                    self.environment_extension,
                    "validation",
                ],
            )
        )
        self.add_environment_logger_with_name_extension("validation")

        parser_logger.debug(
            f"Setup logger. Parser logger has logging level: {self.parser}; "
            f"Environment logger has logging level: {self.environment}"
        )

    def add_environment_logger_with_name_extension(
        self, extension: str
    ) -> logging.Logger:
        """Initializes a logger with the settings for the environment logger.

        The name of the base environment logger is extended by the
        :attr:`extension`. The name if the created logger is found by joining
        the strings of the following variables by an underscore.
        :attr:`Verbosity.base_logger_name`,
        :attr:`Verbosity.environment_extension`, :attr:`extension` Any empty
        strings are ignored.

        Args:
            extension (str):
                Used to differentiate between different environment loggers.
                If empty the base logger is created/returned.

        Returns:
            logging.Logger:
                Logger of the name defined name. The name definition is
                explained in the main documentation of the function.
        """
        logger_name = "_".join(
            filter(
                ("").__ne__,
                [self.base_logger_name, self.environment_extension, extension],
            )
        )
        if not logging.getLogger(logger_name).hasHandlers() or not (
            len(logging.getLogger(logger_name).handlers) > 0
        ):
            logger = set_up_logger(
                logger_name,
                self.logfile_location,
                self.environment,
                self.console_logging,
            )
        else:
            logger = logging.getLogger(logger_name)
        return logger



import datetime
import json
import logging

import numpy as np
import pytest
from conftest import dir_save_location_path

from releso.util.util_funcs import (
    JSONEncoder,
    call_commandline,
    get_path_extension,
    join_infos,
)
from releso.verbosity import VerbosityLevel


@pytest.mark.parametrize(
    "dictionary, wanted_value",
    [
        ({"test": np.ones(12)}, np.ones(12).tolist()),
        ({"test": np.int64(123)}, 123),
        ({"test": bytes("test", "utf-8")}, "test"),
        ({"test": np.longdouble(13.1)}, False),
    ],
)
def test_json_encode(dictionary, wanted_value, capsys):
    with pytest.raises(TypeError):
        json.dumps(dictionary)
    if not wanted_value:
        with pytest.raises(TypeError):
            json.dumps(dictionary, cls=JSONEncoder)
        assert "class 'numpy.longdouble'" in capsys.readouterr().out
        return
    json_string = json.dumps(dictionary, cls=JSONEncoder)
    py_dict = json.loads(json_string)
    assert py_dict["test"] == wanted_value


def test_get_path_extension():
    ret_str = get_path_extension()
    assert datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S") in ret_str


def test_join_infos():
    starting_dict = {"test": 1, "test1": 1}
    new_dict = {"test2": 2, "test": 3}
    join_infos(starting_dict, new_dict, "test")
    assert starting_dict == {"test": 3, "test1": 1, "test2": 2}


def test_call_command_line(dir_save_location, clean_up_provider, caplog):
    dir_save_location_path(True)
    log = logging.getLogger("test")
    log.setLevel(logging.DEBUG)
    (dir_save_location / "test.txt").touch()
    with caplog.at_level(VerbosityLevel.DEBUG):
        exit_code, output = call_commandline("ls", dir_save_location)
        exit_code1, _ = call_commandline(
            "ls",
            ".",
            logging.getLogger("test"),
        )
        assert "Executing command ls in ." in caplog.text
        exit_code2, _ = call_commandline(
            "lsl", ".", logging.getLogger("test")
        )
        assert "Execution failed with return code" in caplog.text
    assert exit_code == 0
    assert exit_code1 == 0
    assert exit_code2 == 127
    assert output == b"test.txt\n"
    clean_up_provider(dir_save_location)



import copy

import numpy as np
import pytest
from pydantic import ValidationError

from releso.shape_parameterization import VariableLocation
from releso.spline import (
    BSplineDefinition,
    NURBSDefinition,
    SplineDefinition,
    SplineSpaceDimension,
)
from releso.util.logger import VerbosityLevel


@pytest.mark.parametrize(
    (
        "n_points",
        "degree",
        "knot_vector",
        "expected_knot_vector",
        "error",
        "msg",
        "verbosity_level",
    ),
    [
        (
            2,
            1,
            None,
            [0, 0, 1, 1],
            False,
            "The knot_vector for dimension test is not given",
            VerbosityLevel.DEBUG,
        ),
        (
            2,
            1,
            [0, 0, 1, 1],
            [0, 0, 1, 1],
            False,
            "The knot_vector for dimension test is given.",
            VerbosityLevel.DEBUG,
        ),
        (
            2,
            None,
            None,
            [0, 0, 1, 1],
            True,
            "During validation the prerequisite",
            None,
        ),
        (
            3,
            1,
            None,
            [0, 0, 0.5, 1, 1],
            False,
            None,
            None,
        ),
        (
            2,
            1,
            [0, 0, 0, 1, 1, 1],
            [0, 0, 0, 1, 1, 1],
            True,
            "The knot vector does not contain the correct number of items",
            VerbosityLevel.DEBUG,
        ),
        (
            3,
            4,
            None,
            [0, 0, 0, 0, 0, 1, 1, 1, 1, 1],
            False,
            "Knot vector is created by adding the starting and ending parts.",
            VerbosityLevel.WARNING,
        ),
        (
            4,
            2,
            None,
            [0, 0, 0, 0.5, 1, 1, 1],
            False,
            None,
            None,
        ),
    ],
)
def test_spline_space_dimension_knot_vector(
    n_points,
    degree,
    knot_vector,
    expected_knot_vector,
    error,
    msg,
    verbosity_level,
    dir_save_location,
    caplog,
):
    calling_dict = {
        "save_location": dir_save_location,
        "name": "test",
        "number_of_points": n_points,
    }
    if degree:
        calling_dict["degree"] = degree
    if knot_vector:
        calling_dict["knot_vector"] = knot_vector

    if error:
        with pytest.raises(ValidationError) as error:
            spline_space_dimension = SplineSpaceDimension(**calling_dict)
        assert msg in str(error.value)
        return
    if msg:
        with caplog.at_level(verbosity_level, logger="ReLeSO_parser"):
            spline_space_dimension = SplineSpaceDimension(**calling_dict)
        assert msg in caplog.text
    else:
        spline_space_dimension = SplineSpaceDimension(**calling_dict)

    assert spline_space_dimension.knot_vector == expected_knot_vector
    assert spline_space_dimension.get_knot_vector() == expected_knot_vector


@pytest.mark.parametrize(
    "space_dimensions, control_points, wanted_control_points, n_points, error",
    [
        (None, None, None, None, True),
        (
            [
                {"name": "test", "number_of_points": 3, "degree": 1},
            ],
            [[21.0], [123.0], [124.0]],
            {
                "current_position": [[21], [123], [124]],
                "min_value": [[21], [123], [124]],
                "max_value": [[21], [123], [124]],
            },
            3,
            False,
        ),
        (
            [
                {"name": "test", "number_of_points": 3, "degree": 1},
            ],
            None,
            {
                "current_position": [[0.0], [0.5], [1.0]],
                "min_value": [[0.0], [0.25], [0.75]],
                "max_value": [[0.25], [0.75], [1.0]],
            },
            3,
            False,
        ),
        (
            [
                {"name": "test", "number_of_points": 3, "degree": 1},
                {"name": "test", "number_of_points": 4, "degree": 1},
            ],
            None,
            {
                "current_position": [
                    [0.0, 0.0],
                    [0.5, 0.0],
                    [1.0, 0.0],
                    [0.0, 1 / 3],
                    [0.5, 1 / 3],
                    [1.0, 1 / 3],
                    [0.0, 2 / 3],
                    [0.5, 2 / 3],
                    [1.0, 2 / 3],
                    [0.0, 1],
                    [0.5, 1],
                    [1.0, 1],
                ],
                "min_value": [
                    [0.0, 0.0],
                    [0.25, 0.0],
                    [0.750, 0.0],
                    [0.0, 1 / 6],
                    [0.25, 1 / 6],
                    [0.75, 1 / 6],
                    [0, 0.5],
                    [0.25, 0.5],
                    [0.750, 0.5],
                    [0.0, 5 / 6],
                    [0.25, 5 / 6],
                    [0.75, 5 / 6],
                ],
                "max_value": [
                    [0.25, 1 / 6],
                    [0.75, 1 / 6],
                    [1, 1 / 6],
                    [0.25, 0.5],
                    [0.75, 0.5],
                    [1, 0.5],
                    [0.25, 5 / 6],
                    [0.75, 5 / 6],
                    [1, 5 / 6],
                    [0.25, 1],
                    [0.75, 1],
                    [1, 1],
                ],
            },
            12,
            False,
        ),
    ],
)
def test_spline_definition_initiate(
    space_dimensions,
    control_points,
    wanted_control_points,
    n_points,
    error,
    dir_save_location,
):
    calling_dict = {
        "save_location": dir_save_location,
    }
    # only add these variables if they are not None
    if space_dimensions:
        calling_dict["space_dimensions"] = space_dimensions
    if control_points:
        calling_dict["control_points"] = control_points
    # check if error is expected
    if error:
        with pytest.raises(ValidationError) as error:
            SplineDefinition(**calling_dict)
        assert " the prerequisite variable space_dimensions" in str(
            error.value
        )
        return
    # initiate the spline definition
    spline_definition = SplineDefinition(**calling_dict)
    for o_var, o_w_cur, o_w_min, o_w_max in zip(
        spline_definition.control_points,
        *[wanted_control_points[key] for key in wanted_control_points.keys()],
    ):
        assert len(o_var) == len(o_w_cur) == len(o_w_min) == len(o_w_max)
        for var, w_cur, w_min, w_max in zip(
            o_var,
            o_w_cur,
            o_w_min,
            o_w_max,
        ):
            assert np.isclose(var.current_position, w_cur)
            assert np.isclose(var.min_value, w_min)
            assert np.isclose(var.max_value, w_max)
    assert spline_definition.get_number_of_points() == n_points


@pytest.mark.parametrize(
    "space_dimensions, control_points",
    [
        (
            [
                {"name": "test", "number_of_points": 3, "degree": 1},
            ],
            None,
        ),
        (
            [
                {"name": "test", "number_of_points": 3, "degree": 1},
            ],
            [[21.0], [123.0], [124.0]],
        ),
        (
            [
                {"name": "test", "number_of_points": 3, "degree": 1},
                {"name": "test", "number_of_points": 4, "degree": 1},
            ],
            None,
        ),
    ],
)
def test_bspline_definition_initiate(
    space_dimensions,
    control_points,
    dir_save_location,
):
    init_dict = {
        "save_location": dir_save_location,
    }
    if space_dimensions:
        init_dict["space_dimensions"] = space_dimensions
    if control_points:
        init_dict["control_points"] = control_points
    spline_def = SplineDefinition(**init_dict)
    bspline_def = BSplineDefinition(**init_dict)
    assert bspline_def.control_points == spline_def.control_points
    assert (
        bspline_def.get_number_of_points() == spline_def.get_number_of_points()
    )
    for bspline_dim, spline_dim in zip(
        bspline_def.space_dimensions, spline_def.space_dimensions
    ):
        assert bspline_dim.name == spline_dim.name
        assert bspline_dim.number_of_points == spline_dim.number_of_points
        assert bspline_dim.degree == spline_dim.degree
        assert bspline_dim.knot_vector == spline_dim.knot_vector

    bspline = bspline_def.get_shape()

    for b_kv, spline_dim in zip(
        bspline.knot_vectors, spline_def.space_dimensions
    ):
        assert np.allclose(b_kv, spline_dim.knot_vector)

    assert np.allclose(bspline.control_points, spline_def.get_parameter_values())


@pytest.mark.parametrize(
    (
        "space_dimensions",
        "control_points",
        "weights",
        "wanted_weights",
        "error",
        "convert",
    ),
    [
        (
            [
                {"name": "test", "number_of_points": 3, "degree": 1},
            ],
            None,
            None,
            None,
            "weights\n  field required",
            False,
        ),
        (
            [
                {"name": "test", "number_of_points": 3, "degree": 1},
            ],
            None,
            [0.5, 0.5, 0.5],
            [0.5, 0.5, 0.5],
            False,
            False,
        ),
        (
            [
                {"name": "test", "number_of_points": 3, "degree": 1},
            ],
            None,
            [0.5, 0.5, 0.5, 0.4],
            None,
            "The length of the weight vector 4 is not the same as the number",
            False,
        ),
        (
            [
                {"name": "test", "number_of_points": 3, "degree": 1},
            ],
            None,
            [0.5, 0.5, 0.4],
            None,
            False,
            True,
        ),
        (
            [
                {"name": "test", "number_of_points": 3, "degree": 1},
            ],
            None,
            ["0.5", "asdsad", 0.4],
            None,
            "value is not a valid float",
            False,
        ),
        # (
        #     [
        #         {"name": "test", "number_of_points": 3, "degree": 1},
        #     ],
        #     [[21.0], [123.0], [124.0]],
        #     None,
        #     None,
        #     False,
        # ),
        # (
        #     [
        #         {"name": "test", "number_of_points": 3, "degree": 1},
        #         {"name": "test", "number_of_points": 4, "degree": 1},
        #     ],
        #     None,
        #     None,
        #     None,
        #     False,
        # ),
    ],
)
def test_nurbs_definition_initiate(
    space_dimensions,
    control_points,
    weights,
    wanted_weights,
    error,
    convert,
    dir_save_location,
):
    if convert:
        if weights:
            weights = [
                VariableLocation(
                    current_position=element,
                    min_value=element - 1,
                    max_value=element,
                    save_location=dir_save_location,
                )
                for element in weights
            ]
    init_dict = {
        "save_location": dir_save_location,
    }
    if space_dimensions:
        init_dict["space_dimensions"] = space_dimensions
    if control_points:
        init_dict["control_points"] = control_points
    spline_def = SplineDefinition(**copy.deepcopy(init_dict))
    if weights:
        init_dict["weights"] = weights
    if error:
        with pytest.raises(ValidationError) as err:
            NURBSDefinition(**init_dict)
        assert error in str(err.value)
        return
    nurbs_def = NURBSDefinition(**init_dict)
    assert nurbs_def.control_points == spline_def.control_points
    assert (
        nurbs_def.get_number_of_points() == spline_def.get_number_of_points()
    )
    for nurbs_dim, spline_dim in zip(
        nurbs_def.space_dimensions, spline_def.space_dimensions
    ):
        assert nurbs_dim.name == spline_dim.name
        assert nurbs_dim.number_of_points == spline_dim.number_of_points
        assert nurbs_dim.degree == spline_dim.degree
        assert nurbs_dim.knot_vector == spline_dim.knot_vector

    nurbs = nurbs_def.get_shape()

    for b_kv, spline_dim in zip(
        nurbs.knot_vectors, spline_def.space_dimensions
    ):
        assert np.allclose(b_kv, spline_dim.knot_vector)

    assert np.allclose(nurbs.control_points, spline_def.get_parameter_values())
    actions = nurbs_def.get_actions()
    assert len(actions) == (6 if convert else 3)
    original_actions = copy.deepcopy(actions)
    for action in actions:
        action.apply_discrete_action(False)
    assert not np.allclose(
        [action.current_position for action in actions],
        [action.current_position for action in original_actions],
    )
    nurbs_def.reset()
    assert np.allclose(
        [action.current_position for action in actions],
        [action.current_position for action in original_actions],
    )



"""ReLeSO logging.

This file holds the verbosity definitions and helper functions for the logger
used in this package.
"""

import enum
import logging
import pathlib
from typing import Optional


def get_parser_logger() -> logging.Logger:
    """Get the logger which is to be used during the parsing of the json file.

    Returns:
        logging.Logger: requested logger
    """
    return logging.getLogger("ReLeSO_parser")


class VerbosityLevel(enum.IntEnum):
    """Defines the verbosity level of the logger."""

    #: Writes only errors to the log
    ERROR = logging.ERROR
    #: Writes errors and warning to the log
    WARNING = logging.WARNING
    #: Writes errors, warnings and info fields to the log
    INFO = logging.INFO
    #: Writes everything to the log.
    DEBUG = logging.DEBUG


def set_up_logger(
    logger_name: str = "",
    log_file_location: pathlib.Path = pathlib.Path("."),
    verbosity: VerbosityLevel = VerbosityLevel.INFO,
    console_logging: bool = False,
    logger: Optional[logging.Logger] = None,
) -> logging.Logger:
    """Create a logger instance with a specified name.

    Author:
        Daniel Wolff (wolff@cats.rwth-aachen.de),
        Clemens Fricke (clemens.david.fricke@tuwien.ac.at)

    Args:
        logger_name (str, optional):
            Name of the logging instance. Defaults to ''.
        log_file_location (Path):
            Path to the directory into which the log file(s) will be placed
            into. Defaults to ".".
        verbosity (VerbosityLevel):
            Enum value for the verbosity of the given logger.
            Defaults to VerbosityLevel.INFO
        console_logging (bool):
            Toggle whether to also log into the console. Defaults to False.
        logger (Optional[logging.Logger]):
            Will add handles defined in this function to the given logger.
            Instead of creating a new logger instance. Defaults to None.

    Returns:
        logging.Logger:
            Configured logger instance for simultaneously writing to file and
            console
    """
    if not logger:  # if no logger is given a new logger is created
        logger = logging.getLogger(logger_name)
    # with existing loggers the wanted logger name is added to the existing
    # name. (multiprocessing)
    else:
        logger.name += f"_{logger_name}"
    # delete old handlers
    logger.handlers.clear()
    # log everything which is debug or above
    logger.setLevel(verbosity.value)
    # create formatter for file output and add it to the handlers
    file_formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )

    log_file_location.mkdir(parents=True, exist_ok=True)

    if verbosity <= VerbosityLevel.WARNING:
        log_file_name = log_file_location / f"{logger_name}.log"
        # create log handler which logs even debug messages
        lh = logging.FileHandler(log_file_name)
        # create console handler with a higher log level
        lh.setLevel(verbosity.value)
        lh.setFormatter(file_formatter)
        logger.addHandler(lh)

    if console_logging:
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        # create formatter for console output and add it to the handlers
        console_formatter = logging.Formatter("%(levelname)s - %(message)s")
        ch.setFormatter(console_formatter)
        logger.addHandler(ch)

    # create error handler which logs only error messages
    err_file_name = log_file_location / f"{logger_name}.err"
    eh = logging.FileHandler(err_file_name)
    eh.setLevel(logging.ERROR)
    eh.setFormatter(file_formatter)
    # add the handlers to logger
    logger.addHandler(eh)

    return logger
